#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

#cython: language_level=3

cimport cython
from libcpp.vector cimport vector

import numpy as np
cimport numpy as np

cdef extern from "nu_fowler.h" namespace "Numina":
    cdef cppclass FowlerResult[T]:
        FowlerResult() except +
        T value
        T variance
        char npix
        char mask

# Generic, valid with teff > 0
    FowlerResult[double] axis_fowler(vector[double] buff, double teff, double gain, double ron, double ts, double blank)
# RON limited case
    FowlerResult[double] axis_fowler_ron(vector[double] buff, double teff, double gain, double ron, double ts, double blank)
    ctypedef FowlerResult[double] (*axis_fowler_func_t)(vector[double] buff, double teff, double gain, double ron, double ts, double blank)

@cython.boundscheck(False)
@cython.wraparound(False)
def _process_fowler_intl(datacube_t arr, double tint, double ts, double gain, double ron, mask_t badpix, double saturation, double blank,
        result_t res, 
        result_t var, 
        mask_t npix,
        mask_t mask
        ):
    '''Loop over the first axis applying Fowler processing.'''
    cdef:
        size_t xr = arr.shape[2]
        size_t yr = arr.shape[1]
        size_t zr = arr.shape[0]
        size_t x, y, z
        double teff
        size_t np = zr // 2
        FowlerResult[double] fres
        double val1, val2
        char bp 
        vector[double] buff
        axis_fowler_func_t axis_func = axis_fowler

    # integration time minus sample times number of pairs
    teff = tint - (np - 1) * ts
    if teff <= 0:
        # where are in a RON limited case
        axis_func = axis_fowler_ron

    buff.reserve(zr)

    for x in range(xr):
        for y in range(yr):
            bp = badpix[y, x]
            if bp == MASK_GOOD:
                for z in range(np):
                    val1 = arr[z,y,x]
                    val2 = arr[z+np,y,x]
                    if val1 < saturation and val2 < saturation:
                        buff.push_back(val2 - val1)

                fres = axis_func(buff, teff, gain, ron, ts, blank)
            else:
                fres.value = fres.variance = blank
                fres.npix = 0
                fres.mask = bp

            res[y,x] = fres.value
            var[y,x] = fres.variance
            npix[y,x] = fres.npix
            mask[y,x] = fres.mask
            buff.clear()

    return res, var, npix, mask




#
# Copyright 2017-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Load RecipeResults from GTC's result.json files"""

import os.path

import numina.util.objimport as objimp
import numina.types.dataframe as dataframe


def process_node(node):
    """Process a node in result.json structure"""
    value = node['value']
    mname = node['name']
    typeid = node['typeid']
    if typeid == 52:  # StructDataValue
        obj = {}
        for el in value['elements']:
            key, val = process_node(el)
            obj[key] = val

        if value['struct_type'] != 'dict':
            # Value is not a dict
            klass = objimp.import_object(value['struct_type'])
            newobj = klass.__new__(klass)
            if hasattr(newobj, '__setstate__'):
                newobj.__setstate__(obj)
            else:
                newobj.__dict__ = obj
            obj = newobj
    elif typeid == 9:
        data = value['data']
        # dim = value['dimension']
        # shape = dim['height'], dim['width']
        obj = data
    elif typeid == 90:  # StructDataValueList
        obj = []
        for el in value:
            sobj = {}
            for sel in el['elements']:
                key, val = process_node(sel)
                sobj[key] = val
            obj.append(sobj)

    elif typeid == 45:  # Frame
        obj = dataframe.DataFrame(frame=os.path.abspath(value['path']))
    else:
        obj = value

    return mname, obj


def build_result(data):
    """Create a dictionary with the contents of result.json"""
    more = {}
    for key, value in data.items():
        if key != 'elements':
            newnode = value
        else:
            newnode = {}
            for el in value:
                nkey, nvalue = process_node(el)
                newnode[nkey] = nvalue

        more[key] = newnode

    return more


if __name__ == '__main__':

    import json

    filename = "result.json"

    with open(filename) as fd:
        data = json.load(fd)

    res = build_result(data)

    print(res)



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy as np
from numpy.polynomial import Polynomial

from ..display.matplotlib_qt import set_window_geometry
from ..display.pause_debugplot import pause_debugplot


def find_peaks_spectrum(sx, nwinwidth, threshold=0, debugplot=0):
    """Find peaks in array.

    The algorithm imposes that the signal at both sides of the peak
    decreases monotonically.

    Parameters
    ----------
    sx : 1d numpy array, floats
        Input array.
    nwinwidth : int
        Width of the window where each peak must be found.
    threshold : float
        Minimum signal in the peaks.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed:
        00 : no debug, no plots
        01 : no debug, plots without pauses
        02 : no debug, plots with pauses
        10 : debug, no plots
        11 : debug, plots without pauses
        12 : debug, plots with pauses

    Returns
    -------
    ixpeaks : 1d numpy array, int
        Peak locations, in array coordinates (integers).

    """

    if not isinstance(sx, np.ndarray):
        raise ValueError("sx=" + str(sx) + " must be a numpy.ndarray")
    elif sx.ndim != 1:
        raise ValueError("sx.ndim=" + str(sx.ndim) + " must be 1")

    sx_shape = sx.shape
    nmed = nwinwidth//2

    if debugplot >= 10:
        print('find_peaks_spectrum> sx shape......:', sx_shape)
        print('find_peaks_spectrum> nwinwidth.....:', nwinwidth)
        print('find_peaks_spectrum> nmed..........:', nmed)
        print('find_peaks_spectrum> data_threshold:', threshold)
        print('find_peaks_spectrum> the first and last', nmed,
              'pixels will be ignored')

    xpeaks = []  # list to store the peaks

    if sx_shape[0] < nwinwidth:
        print('find_peaks_spectrum> sx shape......:', sx_shape)
        print('find_peaks_spectrum> nwinwidth.....:', nwinwidth)
        raise ValueError('sx.shape < nwinwidth')

    i = nmed
    while i < sx_shape[0] - nmed:
        if sx[i] > threshold:
            peak_ok = True
            j = 0
            loop = True
            while loop:
                if sx[i - nmed + j] > sx[i - nmed + j + 1]:
                    peak_ok = False
                j += 1
                loop = (j < nmed) and peak_ok
            if peak_ok:
                j = nmed + 1
                loop = True
                while loop:
                    if sx[i - nmed + j - 1] < sx[i - nmed + j]:
                        peak_ok = False
                    j += 1
                    loop = (j < nwinwidth) and peak_ok
            if peak_ok:
                xpeaks.append(i)
                i += nwinwidth - 1
            else:
                i += 1
        else:
            i += 1

    ixpeaks = np.array(xpeaks)

    if debugplot >= 10:
        print('find_peaks_spectrum> number of peaks found:', len(ixpeaks))
        print(ixpeaks)

    return ixpeaks


def refine_peaks_spectrum(sx, ixpeaks, nwinwidth, method=None,
                          geometry=None, debugplot=0):
    """Refine line peaks in spectrum.

    Parameters
    ----------
    sx : 1d numpy array, floats
        Input array.
    ixpeaks : 1d numpy array, int
        Initial peak locations, in array coordinates (integers).
        These values can be the output from the function
        find_peaks_spectrum().
    nwinwidth : int
        Width of the window where each peak must be refined.
    method : string
        "poly2" : fit to a 2nd order polynomial
        "gaussian" : fit to a Gaussian
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed:
        00 : no debug, no plots
        01 : no debug, plots without pauses
        02 : no debug, plots with pauses
        10 : debug, no plots
        11 : debug, plots without pauses
        12 : debug, plots with pauses

    Returns
    -------
    fxpeaks : 1d numpy array, float
        Refined peak locations, in array coordinates.
    sxpeaks : 1d numpy array, float
        When fitting Gaussians, this array stores the fitted line
        widths (sigma). Otherwise, this array returns zeros.

    """

    nmed = nwinwidth//2

    xfpeaks = np.zeros(len(ixpeaks))
    sfpeaks = np.zeros(len(ixpeaks))

    for iline in range(len(ixpeaks)):
        jmax = ixpeaks[iline]
        x_fit = np.arange(-nmed, nmed+1, dtype=float)
        # prevent possible problem when fitting a line too near to any
        # of the borders of the spectrum
        j1 = jmax - nmed
        j2 = jmax + nmed + 1
        if j1 < 0:
            j1 = 0
            j2 = 2 * nmed + 1
            if j2 >= len(sx):
                raise ValueError("Unexpected j2=" + str(j2) +
                                 " value when len(sx)=" + str(len(sx)))
        if j2 >= len(sx):
            j2 = len(sx)
            j1 = j2 - (2 * nmed + 1)
            if j1 < 0:
                raise ValueError("Unexpected j1=" + str(j1) +
                                 " value when len(sx)=" + str(len(sx)))
        # it is important to create a copy in the next instruction in
        # order to avoid modifying the original array when normalizing
        # the data to be fitted
        y_fit = np.copy(sx[j1:j2].astype(float))
        sx_peak_flux = y_fit.max()
        if sx_peak_flux != 0:
            y_fit /= sx_peak_flux  # normalize to maximum value

        if method == "gaussian":
            # check that there are no negative or null values
            if y_fit.min() <= 0:
                if debugplot >= 10:
                    print("WARNING: negative or null value encountered" +
                          " in refine_peaks_spectrum with gaussian.")
                    print("         Using poly2 method instead.")
                final_method = "poly2"
            else:
                final_method = "gaussian"
        else:
            final_method = method

        if final_method == "poly2":
            poly_funct = Polynomial.fit(x_fit, y_fit, 2)
            poly_funct = Polynomial.cast(poly_funct)
            coef = poly_funct.coef
            if len(coef) == 3:
                if coef[2] != 0:
                    refined_peak = -coef[1]/(2.0*coef[2]) + jmax
                else:
                    refined_peak = 0.0 + jmax
            else:
                refined_peak = 0.0 + jmax
        elif final_method == "gaussian":
            poly_funct = Polynomial.fit(x_fit, np.log(y_fit), 2)
            poly_funct = Polynomial.cast(poly_funct)
            coef = poly_funct.coef
            if len(coef) == 3:
                if coef[2] != 0:
                    refined_peak = -coef[1]/(2.0*coef[2]) + jmax
                else:
                    refined_peak = 0.0 + jmax
                if coef[2] >= 0:
                    sfpeaks[iline] = None
                else:
                    sfpeaks[iline] = np.sqrt(-1 / (2.0 * coef[2]))
            else:
                refined_peak = 0.0 + jmax
                sfpeaks[iline] = None
        else:
            raise ValueError("Invalid method=" + str(final_method) + " value")

        xfpeaks[iline] = refined_peak

        if debugplot % 10 != 0:
            from numina.array.display.matplotlib_qt import plt
            fig = plt.figure()
            set_window_geometry(geometry)
            ax = fig.add_subplot(111)
            xmin = x_fit.min()-1
            xmax = x_fit.max()+1
            ymin = 0
            ymax = y_fit.max()*1.10
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)
            ax.set_xlabel('index around initial integer peak')
            ax.set_ylabel('Normalized number of counts')
            ax.set_title("Fit to line at array index " + str(jmax) +
                         "\n(method=" + final_method + ")")
            plt.plot(x_fit, y_fit, "bo")
            x_plot = np.linspace(start=-nmed, stop=nmed, num=1000,
                                 dtype=float)
            if final_method == "poly2":
                y_plot = poly_funct(x_plot)
            elif final_method == "gaussian":
                amp = np.exp(coef[0] - coef[1] * coef[1] / (4 * coef[2]))
                x0 = -coef[1] / (2.0 * coef[2])
                sigma = np.sqrt(-1 / (2.0 * coef[2]))
                y_plot = amp * np.exp(-(x_plot - x0)**2 / (2 * sigma**2))
            else:
                raise ValueError("Invalid method=" + str(final_method) +
                                 " value")
            ax.plot(x_plot, y_plot, color="red")
            print('Refined peak location:', refined_peak)
            plt.show(block=False)
            plt.pause(0.001)
            pause_debugplot(debugplot)

    return xfpeaks, sfpeaks



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import math

from ..utils import coor_to_pix
from ._traces import tracing


def trace(arr, x, y, axis=0, background=0.0,
          step=4, hs=1, tol=2, maxdis=2.0, gauss=1):
    """Trace peak in array starting in (x,y).

    Trace a peak feature in an array starting in position (x,y).

    Parameters
    ----------
    arr : array
         A 2D array
    x : float
        x coordinate of the initial position
    y : float
        y coordinate of the initial position
    axis : {0, 1}
           Spatial axis of the array (0 is Y, 1 is X).
    background: float
           Background level
    step : int, optional
           Number of pixels to move (left and rigth)
           in each iteration
    gauss : bint, optional
            If 1 --> Gaussian interpolation method
            if 0 --> Linear interpolation method

    Returns
    -------
    ndarray
        A nx3 array, with x,y,p of each point in the trace
    """

    i, j = coor_to_pix([x, y], order='xy')
    value = arr[i, j]

    # If arr is not in native byte order, the C-extension won't work

    if arr.dtype.byteorder != '=':
        arr2 = arr.byteswap().newbyteorder()
    else:
        arr2 = arr

    if axis == 0:
        arr3 = arr2
    elif axis == 1:
        arr3 = arr2.t
    else:
        raise ValueError("'axis' must be 0 or 1")

    result = tracing(arr3, x, y, value, background=background,
                     step=step, hs=hs, tol=tol, maxdis=maxdis, gauss=gauss)

    if axis == 1:
        # Flip X,Y columns
        return result[:, ::-1]

    return result


def tracing_limits(size, col, step, hs):
    m = col % step
    k0 = ((hs + step - m) / step)
    r0 = int(math.floor(k0))
    xx_start = r0 * step + m
    k1 = ((size - hs - step - m) / step)
    r1 = int(math.ceil(k1))
    xx_end = r1 * step + m
    return xx_start, xx_end


def axis_to_dispaxis(axis):
    """Obtain the dispersion axis from the spatial axis."""
    if axis == 0:
        dispaxis = 1
    elif axis == 1:
        dispaxis = 0
    else:
        raise ValueError("'axis' must be 0 or 1")
    return dispaxis



import pytest

from numina.keydef import extract, KeyDefinition


@pytest.mark.parametrize(
    "path, result",
    [(["PATTRI.wheel", "label"], "U"), (["PATTRI.wheel", "broken"], "B")],
)
def test_extract(pattri_header, pattri_state, path, result):
    key_def = KeyDefinition("FILTER", default="B")
    extract(pattri_header, pattri_state, path, key_def)
    assert pattri_header["FILTER"] == result



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Define a piecewise linear function."""
import numpy as np


def define_piecewise_linear_function(xarray, yarray):
    """Define a piecewise linear function.

    Parameters
    ----------
    xarray : 1D numpy array
        The x coordinates of the points.
    yarray : 1D numpy array
        The y coordinates of the points.

    """
    isort = np.argsort(xarray)
    xfit = xarray[isort]
    yfit = yarray[isort]

    def function(x):
        y = np.interp(x, xfit, yfit)
        return y

    return function





import numina.util.flow as flow


def test_serial_empty_is_id():

    empty_serial = flow.SerialFlow([])

    class Something(object):
        pass

    inp = Something()

    assert inp is empty_serial.run(inp)



#
# Copyright 2017-2019 Universidad Complutense de Madrid
#
# This file is part of Numina DRP
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import inspect
import numina.exceptions as nexcep
import numina.types.datatype as dt


class MultiType(dt.DataType):
    def __init__(self, *args):

        node_type = []
        for obj in args:
            if inspect.isclass(obj):
                obj = obj()
            node_type.append(obj)

        super(MultiType, self).__init__(ptype=None, node_type=node_type)
        self.current_node = None

    def validate(self, obj):
        if self.current_node:
            return self.current_node.validate(obj)
        else:
            faillures = []
            for subtype in self.node_type:
                try:
                    return subtype.validate(obj)
                except nexcep.ValidationError as val_err:
                    faillures.append(val_err)
            else:
                internal_m = tuple(v.args[1] for v in faillures)
                raise nexcep.ValidationError(obj, internal_m)

    def isproduct(self):
        if self.current_node:
            return self.current_node.isproduct()
        return True

    def tag_names(self):
        if self.current_node:
            return self.current_node.tag_names()
        else:
            join = []
            for subtype in self.node_type:
                join.extend(subtype.tag_names())
            return join

    def extract_tags(self, obj):
        """Extract tags from obj"""
        if self.current_node:
            return self.current_node.extract_tags(obj)
        else:
            # FIXME: unknown type
            for subtype in self.node_type:
                return subtype.extract_tags(obj)
            return {}

    def descriptive_name(self):
        start, remain = self.node_type[0], self.node_type[1:]
        build_str = [start.descriptive_name()]
        for x in remain:
            field = f"or {x.descriptive_name()}"
            build_str.append(field)
        return " ".join(build_str)

    def _datatype_load(self, obj):
        faillures = []
        for subtype in self.node_type:
            try:
                return subtype._datatype_load(obj)
            except KeyError:
                faillures.append(subtype)
        else:
            msg = f"types {faillures} cannot load 'obj'"
            raise TypeError(msg)

    def _query_on_dal(self, name, dal, ob, options=None):

        # Results for subtypes
        results = []
        faillures = []
        for subtype in self.node_type:
            try:
                result = subtype.query(name, dal, ob, options=options)
                results.append((subtype, result))
            except nexcep.NoResultFound as notfound:
                faillures.append((subtype, notfound))
        else:
            # Not found
            for subtype, notfound in faillures:
                subtype.on_query_not_found(notfound)
        # Select wich of the results we choose
        if results:
            # Select the first, for the moment
            subtype, result = results[0]
            self.current_node = subtype
            return result
        else:
            raise nexcep.NoResultFound




import numpy

from .. import combine_shapes


def test_combine_shapes():

    value = 100
    shapes = [(100, 100), (100, 100), (20, 130)]

    # refs must go in (z,y,x) order...
    refs = [(33, 22), (12, 78), (5, 5)]
    arrn = [numpy.zeros(shape, dtype='int') for shape in shapes]

    for arr, ref in zip(arrn, refs):
        arr[ref] = value

    finalshape, slices, finalref = combine_shapes(shapes, refs)
    final = numpy.zeros(finalshape, dtype='int')

    assert len(slices) == len(shapes)

    for arr, sl in zip(arrn, slices):
        final[sl] += arr

    assert final.max() == value * len(shapes)
    assert finalshape == (121, 203)


def test_combine_shapes2():

    value1 = 1
    value2 = 1
    shapes = [(40, 99), (103, 100), (100, 109)]

    # refs must go in (z,y,x) order...
    fshape = (128, 182)
    refs = [(33, 22), (12, 78), (5, 5)]
    # refs = [(0, 0), (5, 0), (0, 0)]

    # off = [(ref[0] - refs[0][0], ref[1] - refs[0][1]) for ref in refs]

    arrn = [numpy.zeros(shape, dtype='int') for shape in shapes]

    for arr, ref in zip(arrn, refs):
        arr[:] = value1
        arr[ref] += value2

    finalshape, slices, finalref = combine_shapes(shapes, refs)

    assert finalshape == fshape

    final = numpy.zeros(finalshape, dtype='int')

    for arr, sl in zip(arrn, slices):
        final[sl] += arr

    assert final[finalref] == (value1 + value2) * len(shapes)



import pytest


from ..generic import InstrumentGeneric
from ..property import PropertyEntry, PropertyModOneOf, PropertyModLimits


@pytest.mark.parametrize("mode, value", [("A", 1), ("B", 2)])
def test_property_entry(mode, value):
    values = {"A": 1, "B": 2}
    depends = ["mode"]
    prop = PropertyEntry(values, depends)
    state = {"mode": mode}
    res = prop.get(**state)
    assert res == value


def test_property_entry_set():
    """Setting this type of property does nothing"""
    values = {"A": 1, "B": 2}
    depends = ["mode"]
    prop1 = PropertyEntry(values, depends)
    prop2 = PropertyModOneOf(["A", "B"], default="A")
    props = {"p1": prop1, "mode": prop2}
    obj = InstrumentGeneric.from_component("B", "b", properties=props)
    assert obj.p1 == 1
    obj.p1 = 2
    assert obj.p1 == 1


def test_property_limit_except():
    with pytest.raises(ValueError):
        prop1 = PropertyModLimits([3])  # noqa


@pytest.mark.parametrize("value, inside", [(-2, False), (0, True), (2, False)])
def test_property_limit_check(value, inside):
    prop1 = PropertyModLimits([-1, 1])
    assert prop1.value_check(value) == inside



#
# -*- coding: utf8 -*-
#
# Copyright 2014-2020 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

'''Mode estimators.'''

import math

import numpy as np


# HSM method
# http://www.davidbickel.com/
def mode_half_sample(a, is_sorted=False):
    """
    Estimate the mode using the Half Sample mode.

    A method to estimate the mode, as described in
    D. R. Bickel and R. Frühwirth (contributed equally),
    "On a fast, robust estimator of the mode: Comparisons to other
    robust estimators with applications,"
    Computational Statistics and Data Analysis 50, 3500-3530 (2006).


    Example
    =======

    >>> import numpy as np
    >>> np.random.seed(1392838)
    >>> a = np.random.normal(1000, 200, size=1000)
    >>> a[:100] = np.random.normal(2000, 300, size=100)
    >>> b = np.sort(a)
    >>> mode_half_sample(b, is_sorted=True)
    1041.9327885039545

    """

    a = np.asanyarray(a)

    if not is_sorted:
        sdata = np.sort(a)
    else:
        sdata = a

    n = len(sdata)
    if n == 1:
        return sdata[0]
    elif n == 2:
        return 0.5 * (sdata[0] + sdata[1])
    elif n == 3:
        ind = -sdata[0] + 2 * sdata[1] - sdata[2]
        if ind < 0:
            return 0.5 * (sdata[0] + sdata[1])
        elif ind > 0:
            return 0.5 * (sdata[1] + sdata[2])
        else:
            return sdata[1]
    else:
        N = int(math.ceil(n / 2.0))
        w = sdata[(N-1):] - sdata[:(n-N+1)]
        ar = w.argmin()
        return mode_half_sample(sdata[ar:ar+N], is_sorted=True)


def mode_sex(a):
    '''Estimate the mode as sextractor'''
    return 2.5 * np.median(a) - 1.5 * np.mean(a)



#
# Copyright 2010-2020 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


try:
    from functools import singledispatch
except ImportError:
    from pkgutil import simplegeneric as singledispatch


@singledispatch
def load(tag, obj):
    """

    Parameters
    ----------
    tag
    obj

    Returns
    -------

    """

    if hasattr(tag, '__numina_load__'):
        return tag.__numina_load__(obj)

    if hasattr(tag, '_datatype_load'):
        return tag._datatype_load(obj)

    return obj



#
# Copyright 2011-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Import objects by name"""

import importlib
import warnings


def import_object(path):
    """Import an object given its fully qualified name."""
    spl = path.split('.')
    if len(spl) == 1:
        return importlib.import_module(path)
    # avoid last part for the moment
    cls = spl[-1]
    mods = '.'.join(spl[:-1])

    mm = importlib.import_module(mods)
    # try to get the last part as an attribute
    try:
        obj = getattr(mm, cls)
        return obj
    except AttributeError:
        pass

    # Try to import the last part
    rr = importlib.import_module(path)
    return rr


def fully_qualified_name(obj, sep='.'):

    warnings.warn(
        "use numina.util.fqn.fully_qualified_name instead",
        DeprecationWarning, stacklevel=2
    )
    import numina.util.fqn as fqn

    return fqn.fully_qualified_name(obj, sep)



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.io import fits
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import os

from .raise_valueerror import raise_ValueError


def simulate_image2d_from_fitsfile(
        infile,
        diagonal_fov_arcsec,
        plate_scale_x,
        plate_scale_y,
        nphotons,
        rng,
        background_to_subtract=None,
        image_threshold=0.0,
        plots=False,
        verbose=False
):
    """Simulate photons mimicking a 2D image from FITS file.

    Parameters
    ----------
    infile : str
        Input file containing the FITS image to be simulated.
    diagonal_fov_arcsec : `~astropy.units.Quantity`
        Desired field of View (arcsec) corresponding to the diagonal
        of the FITS image.
    plate_scale_x : `~astropy.units.Quantity`
        Plate scale of the IFU in the X direction.
    plate_scale_y : `~astropy.units.Quantity`
        Plate scale of the IFU in the Y direction.
    nphotons : int
        Number of photons to be simulated.
    rng : `~numpy.random._generator.Generator`
        Random number generator.
    background_to_subtract : str or None
        If not None, this parameters indicates how to compute
        the background to be subtracted.
    image_threshold : float
        Data below this threshold is set to zero.
    plots : bool
        If True, plot intermediate results.
    verbose : bool
        If True, display additional information.

    Returns
    -------
    simulated_x_ifu : `~numpy.ndarray`
        Array of simulated photon X coordinates in the IFU.
    simulated_y_ify : `~numpy.ndarray`
        Array of simulated photon Y coordinates in the IFU.

    """

    # read input FITS file
    if verbose:
        print(f'Reading {infile=}')
    with fits.open(infile) as hdul:
        image2d_ini = hdul[0].data
    image2d_reference = image2d_ini.astype(float)
    naxis2, naxis1 = image2d_reference.shape
    npixels = naxis1 * naxis2

    # subtract background
    if background_to_subtract == 'mode':
        nbins = int(np.sqrt(npixels) + 0.5)
        h, bin_edges = np.histogram(image2d_reference.flatten(), bins=nbins)
        imax = np.argmax(h)
        skylevel = (bin_edges[imax] + bin_edges[imax+1]) / 2
        if verbose:
            print(f'Subtracting {skylevel=} (image mode)')
    elif background_to_subtract == 'median':
        skylevel = np.median(image2d_reference.flatten())
        if verbose:
            print(f'Subtracting {skylevel=} (image median)')
    elif background_to_subtract == 'none':
        skylevel = 0.0
        if verbose:
            print('Skipping background subtraction')
    else:
        try:
            skylevel = float(background_to_subtract)
        except ValueError:
            skylevel = None   # avoid PyCharm warning (not aware of raise ValueError)
            raise_ValueError(f'Invalid {background_to_subtract=}')
        if verbose:
            print(f"Subtracting {skylevel=} (user's value)")
    image2d_reference -= skylevel

    # impose image threshold
    if verbose:
        print(f'Applying {image_threshold=}')
    image2d_reference[image2d_reference <= image_threshold] = 0
    if np.min(image2d_reference) < 0.0:
        raise_ValueError(f'{np.min(image2d_reference)=} must be >= 0.0')

    # flatten image to be simulated
    image1d = image2d_reference.flatten()
    # compute normalized cumulative area
    xpixel = 1 + np.arange(npixels)
    cumulative_area = np.concatenate((
        [0],
        np.cumsum((image1d[:-1] + image1d[1:]) / 2 * (xpixel[1:] - xpixel[:-1]))
    ))
    normalized_cumulative_area = cumulative_area / cumulative_area[-1]
    if plots:
        fig, ax = plt.subplots()
        ax.plot(xpixel, normalized_cumulative_area, '.')
        ax.set_xlabel(f'xpixel')
        ax.set_ylabel('Normalized cumulative area')
        ax.set_title(os.path.basename(infile))
        plt.tight_layout()
        plt.show()
    # invert normalized cumulative area using random uniform samples
    unisamples = rng.uniform(low=0, high=1, size=nphotons)
    simulated_pixel = np.interp(x=unisamples, xp=normalized_cumulative_area, fp=xpixel)
    # compute histogram of 1D data
    bins_pixel = 0.5 + np.arange(npixels + 1)
    int_simulated_pixel, bin_edges = np.histogram(simulated_pixel, bins=bins_pixel)
    # reshape 1D into 2D image
    image2d_simulated = int_simulated_pixel.reshape((naxis2, naxis1))
    # scale factors to insert simulated image in requested field of view
    plate_scale = diagonal_fov_arcsec / (np.sqrt(naxis1**2 + naxis2**2) * u.pix)
    factor_x = abs((plate_scale / plate_scale_x.to(u.arcsec / u.pix)).value)
    factor_y = abs((plate_scale / plate_scale_y.to(u.arcsec / u.pix)).value)
    # redistribute photons in each pixel of the simulated image using a
    # random distribution within the considered pixel
    jcenter = naxis1 / 2
    icenter = naxis2 / 2
    simulated_x_ifu = []
    simulated_y_ifu = []
    for i, j in np.ndindex(naxis2, naxis1):
        nphotons_in_pixel = image2d_simulated[i, j]
        if nphotons_in_pixel > 0:
            jmin = j - jcenter - 0.5
            jmax = j - jcenter + 0.5
            simulated_x_ifu += (rng.uniform(low=jmin, high=jmax, size=nphotons_in_pixel) * factor_x).tolist()
            imin = i - icenter - 0.5
            imax = i - icenter + 0.5
            simulated_y_ifu += (rng.uniform(low=imin, high=imax, size=nphotons_in_pixel) * factor_y).tolist()

    # return result
    return np.array(simulated_x_ifu), np.array(simulated_y_ifu)



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy as np

from .matplotlib_qt import set_window_geometry
from .pause_debugplot import pause_debugplot


def ximplot(ycut, title=None, show=True, plot_bbox=(0, 0),
            geometry=(0, 0, 640, 480), tight_layout=True,
            debugplot=None):
    """Auxiliary function to display 1d plot.

    Parameters
    ----------
    ycut : 1d numpy array, float
        Array to be displayed.
    title : string
        Plot title.
    show : bool
        If True, the function shows the displayed image. Otherwise
        plt.show() is expected to be executed outside.
    plot_bbox : tuple (2 integers)
        If tuple is (0,0), the plot is displayed with image
        coordinates (indices corresponding to the numpy array).
        Otherwise, the bounding box of the image is read from this
        tuple, assuming (nc1,nc2). In this case, the coordinates
        indicate pixels.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    tight_layout : bool
        If True, and show=True, a tight display layout is set.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    ax : axes object
        Matplotlib axes instance. This value is returned only when
        'show' is False.

    """

    # protections
    if type(ycut) is not np.ndarray:
        raise ValueError("ycut=" + str(ycut) +
                         " must be a numpy.ndarray")
    elif ycut.ndim != 1:
        raise ValueError("ycut.ndim=" + str(ycut.dim) +
                         " must be 1")

    # read bounding box limits
    nc1, nc2 = plot_bbox
    plot_coord = (nc1 == 0 and nc2 == 0)

    naxis1_ = ycut.size
    if not plot_coord:
        # check that ycut size corresponds to bounding box size
        if naxis1_ != nc2 - nc1 + 1:
            raise ValueError("ycut.size=" + str(ycut.size) +
                             " does not correspond to bounding box size")

    # display image
    from numina.array.display.matplotlib_qt import plt
    if not show:
        plt.ioff()

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.autoscale(False)
    ymin = ycut.min()
    ymax = ycut.max()
    if plot_coord:
        xmin = -0.5
        xmax = (naxis1_ - 1) + 0.5
        xcut = np.arange(naxis1_, dtype=float)
        ax.set_xlabel('image array index in the X direction')
        ax.set_ylabel('pixel value')
    else:
        xmin = float(nc1) - 0.5
        xmax = float(nc2) + 0.5
        xcut = np.linspace(start=nc1, stop=nc2, num=nc2 - nc1 + 1)
        ax.set_xlabel('image pixel in the X direction')
        ax.set_ylabel('pixel value')
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    ax.plot(xcut, ycut, '-')
    if title is not None:
        ax.set_title(title)

    # set the geometry
    set_window_geometry(geometry)

    if show:
        pause_debugplot(debugplot, pltshow=show, tight_layout=tight_layout)
    else:
        if tight_layout:
            plt.tight_layout()
        # return axes
        return ax




import pytest

import numina.tests.simpleobj as simple


@pytest.mark.result_compare
def test_result_compare_demo1():
    from numina.types.frame import DataFrameType
    from numina.types.structured import BaseStructuredCalibration

    import numina.core

    simple_img = simple.create_simple_hdul()
    simple_stc = simple.create_simple_structured()

    class BB(numina.core.RecipeResult):
        prod1 = numina.core.Result(int, 'something1')
        prod2 = numina.core.Result(DataFrameType, 'something2')
        prod3 = numina.core.Result(BaseStructuredCalibration, 'something3')

    result = BB(prod1=1, prod2=simple_img, prod3=simple_stc)
    return result


@pytest.mark.result_compare
def test_result_compare_demo2():
    from numina.types.frame import DataFrameType
    from numina.types.structured import BaseStructuredCalibration

    import numina.core

    simple_img = simple.create_simple_hdul()
    simple_stc = simple.create_simple_structured()

    class BB(numina.core.RecipeResult):
        prod1 = numina.core.Result(int, description='something1',
                                   destination="prod4")
        prod2 = numina.core.Result(DataFrameType, 'something2')
        prod3 = numina.core.Result(BaseStructuredCalibration, 'something3')

    result = BB(prod1=1, prod2=simple_img, prod3=simple_stc)
    return result



# Copyright 2017-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import json
import uuid
import datetime

import builtins

import numina.types.product
import numina.types.datatype
from numina.ext.gtc import DF
import numina.util.convert as conv
from numina.util.jsonencoder import ExtEncoder
from numina.util.fqn import fully_qualified_name
from numina.util.objimport import import_object


def writeto(obj, name):
    """
    Write a BaseStructuredCalibration to a named file

    Parameters
    ----------
    obj : BaseStructuredCalibration
    name : str or file or file-like or pathlib.Path


    Returns
    -------

    """
    with builtins.open(name, 'w') as fd:
        json.dump(obj.__getstate__(), fd, indent=2, cls=ExtEncoder)


def open(name):
    """"
    Open a JSON file as a structured object

    name : str or file or file-like or pathlib.Path
        File to be opened
    """
    with builtins.open(name, mode='r') as fd:
        data = json.load(fd)
    return loads(data)


def load(fd):
    """"
    Load a JSON file as a structured object

    fd: file or file-like
      File to be opened
    """
    data = json.load(fd)
    return loads(data)


def loads(data):
    # FIXME: duplicated
    if 'type_fqn' in data:
        type_fqn = data['type_fqn']
        cls = import_object(type_fqn)
        obj = cls.__new__(cls)
        obj.__setstate__(data)
        return obj
    return data


class BaseStructuredCalibration(numina.types.product.DataProductMixin,
                                numina.types.datatype.AutoDataType):
    """Base class for structured calibration data

    Parameters
    ----------

    instrument: str
        Instrument name

    Attributes
    ----------
    tags: dict
        dictionary of selection fields
    uuid: str
       UUID of the result

    """

    def __init__(self, instrument='unknown', datamodel=None):
        super(BaseStructuredCalibration, self).__init__(datamodel=datamodel)
        self.instrument = instrument
        self.tags = {}
        self.uuid = str(uuid.uuid1())

        self.meta_info = self.create_meta_info()
        self.meta_info['instrument_name'] = self.instrument
        self.meta_info['creation_date'] = datetime.datetime.now(
            datetime.timezone.utc).isoformat()
        self._base_info = {}
        self.add_dialect_info('gtc', DF.TYPE_STRUCT)

    @property
    def calibid(self):
        return f'uuid:{self.uuid}'

    @property
    def default(self):
        return None

    def __getstate__(self):

        st = super(BaseStructuredCalibration, self).__getstate__()

        keys = ['instrument', 'tags', 'uuid', 'meta_info']
        for key in keys:
            st[key] = self.__dict__[key]

        st['type'] = self.name()
        st['type_fqn'] = fully_qualified_name(self)
        return st

    def __setstate__(self, state):
        super(BaseStructuredCalibration, self).__setstate__(state)
        # self.add_dialect_info('gtc', DF.TYPE_STRUCT)

        self.instrument = state['instrument']
        self.tags = state['tags']
        self.uuid = state['uuid']
        self.meta_info = {}
        self._base_info = {}
        for key in state:
            if key not in ['contents', 'quality_control']:
                setattr(self, key, state[key])

    def __str__(self):
        sclass = type(self).__name__
        if self.instrument != 'unknown':
            return f"{sclass}(instrument={self.instrument}, uuid={self.uuid})"
        else:
            return f"{sclass}()"

    def writeto(self, name):
        return writeto(self, name)

    @classmethod
    def _datatype_dump(cls, obj, where):

        filename = where + '.json'
        writeto(obj, filename)
        return filename

    @classmethod
    def _datatype_load(cls, obj):
        try:
            with builtins.open(obj, 'r') as fd:
                state = json.load(fd)
        except IOError as e:
            raise e

        result = cls.__new__(cls)
        result.__setstate__(state=state)
        return result

    @staticmethod
    def create_meta_info():
        meta_info = {}
        meta_info['mode_name'] = 'unknown'
        meta_info['instrument_name'] = 'unknown'
        meta_info['recipe_name'] = 'unknown'
        meta_info['recipe_version'] = 'unknown'
        meta_info['origin'] = {}
        return meta_info

    def extract_tags(self, obj):
        """Extract tags"""
        return obj.tags

    def extract_db_info(self, obj, keys):
        """Extract metadata from serialized file"""

        objl = self.convert_in(obj)

        # FIXME: this is too complex
        if isinstance(objl, self.__class__):
            return objl.update_meta_info()

        try:
            with builtins.open(objl, mode='r') as fd:
                state = json.load(fd)
        except IOError as e:
            raise e

        result = super(BaseStructuredCalibration,
                       self).extract_db_info(state, keys)

        try:
            minfo = state['meta_info']
            result['mode'] = minfo['mode_name']
            origin = minfo['origin']
            date_obs = origin['date_obs']
        except KeyError:
            origin = {}
            date_obs = "1970-01-01T00:00:00.00"

        result['instrument'] = state['instrument']
        result['uuid'] = state['uuid']
        result['tags'] = state['tags']
        result['type'] = state['type']
        result['observation_date'] = conv.convert_date(date_obs)
        result['origin'] = origin

        return result

    def update_meta_info(self):
        """Extract metadata from myself"""
        result = super(BaseStructuredCalibration, self).update_meta_info()

        result['instrument'] = self.instrument
        result['uuid'] = self.uuid
        result['tags'] = self.tags
        result['type'] = self.name()

        minfo = self.meta_info
        try:
            result['mode'] = minfo['mode_name']
            origin = minfo['origin']
            date_obs = origin['date_obs']
        except KeyError:
            origin = {}
            date_obs = "1970-01-01T00:00:00.00"

        result['observation_date'] = conv.convert_date(date_obs)
        result['origin'] = origin

        return result

    @property
    def meta(self):
        self._base_info = self.update_meta_info()
        return self._base_info

    def update_metadata(self, recipe):
        try:
            self.meta_info['mode_name'] = recipe.mode.key
        except AttributeError:
            self.meta_info['mode_name'] = recipe.mode

        self.meta_info['instrument_name'] = recipe.instrument
        self.meta_info['recipe_name'] = recipe.__class__.__name__
        self.meta_info['recipe_version'] = recipe.__version__

    def update_metadata_origin(self, obresult_meta):
        origin = self.meta_info['origin']
        origin['block_uuid'] = obresult_meta['block_uuid']
        origin['insconf_uuid'] = obresult_meta['insconf_uuid']
        origin['date_obs'] = obresult_meta['observation_date']
        origin['observation_date'] = obresult_meta['observation_date']
        origin['frames'] = obresult_meta['frames']



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Valid parameters for cosmic ray masks computation and application."""

VALID_LACOSMIC_CLEANTYPE = ['median', 'medmask', 'meanmask', 'idw']
VALID_CRMETHODS = ['mmcosmic', 'lacosmic', 'mm_lacosmic']
VALID_BOUNDARY_FITS = ['spline', 'piecewise']
VALID_COMBINATIONS = ['mediancr', 'meancrt', 'meancr']



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Calculate elliptical 2D Gaussian kernel."""
import math
import numpy as np


def gausskernel2d_elliptical(fwhm_x, fwhm_y, kernsize):
    """Calculate elliptical 2D Gaussian kernel.

    This kernel can be used as the psfk parameter of
    ccdproc.cosmicray_lacosmic.

    Parameters
    ----------
    fwhm_x : float
        Full width at half maximum in the x direction.
    fwhm_y : float
        Full width at half maximum in the y direction.
    kernsize : int
        Size of the kernel (must be odd).

    Returns
    -------
    kernel : 2D numpy array
        The elliptical Gaussian kernel. It is normalized
        so that the sum of all its elements is 1. It
        is returned as a float32 array (required by
        ccdproc.cosmicray_lacosmic).
    """

    if kernsize % 2 == 0 or kernsize < 3:
        raise ValueError("kernsize must be an odd integer >= 3.")
    if fwhm_x <= 0 or fwhm_y <= 0:
        raise ValueError("fwhm_x and fwhm_y must be positive numbers.")

    sigma_x = fwhm_x / (2 * math.sqrt(2 * math.log(2)))
    sigma_y = fwhm_y / (2 * math.sqrt(2 * math.log(2)))
    halfsize = kernsize // 2
    y, x = np.mgrid[-halfsize:halfsize+1, -halfsize:halfsize+1]
    kernel = np.exp(-0.5 * ((x / sigma_x) ** 2 + (y / sigma_y) ** 2))
    kernel /= np.sum(kernel)

    # reverse the psf kernel as that is what it is used in the convolution
    kernel = kernel[::-1, ::-1]
    return kernel.astype(np.float32)



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Unit test for logger"""

import logging

from ..logger import FITSHistoryHandler, log_to_history

import pytest
from astropy.io import fits

from numina.types.dataframe import DataFrame


@pytest.fixture(scope="function")
def logger_fits(request):

    logger = logging.getLogger('numina_test_logger')
    logger.setLevel(logging.DEBUG)
    history_header = fits.Header()
    fh = FITSHistoryHandler(history_header)
    fh.setLevel(logging.INFO)
    logger.addHandler(fh)

    def fin():
        logger.removeHandler(fh)

    request.addfinalizer(fin)

    return logger, history_header  # provide the fixture value


def test_fits_history_handler(logger_fits):

    logger, history_header = logger_fits
    logtext1 = 'Test1'
    logger.info(logtext1)
    logtext2 = 100 * 't'
    logger.info(logtext2)
    hheaders = history_header['HISTORY']

    assert hheaders[0] == logtext1
    assert hheaders[1] == logtext2[:72]
    assert hheaders[2] == logtext2[72:]


def test_logger_decorator():

    _logger = logging.getLogger('numina_test_logger')

    class RecipeResult:
        pass

    class Recipe:
        @log_to_history(_logger, "some")
        def decorated_method(self, ri):
            rr = RecipeResult()
            img0 = fits.PrimaryHDU()
            _logger.info("some")
            _logger.info("other")
            img = fits.HDUList([img0])
            some = DataFrame(frame=img)
            rr.some = some
            return rr

    recipe = Recipe()

    rr = recipe.decorated_method(None)
    h = rr.some.open()[0].header
    assert h['history'][0] == 'some'
    assert h['history'][1] == 'other'



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Compute world_to_pixel.
"""
import argparse
from argparse import RawTextHelpFormatter
from astropy.coordinates import SkyCoord
from astropy.units import Unit
from astropy.io import fits
from astropy.wcs import WCS
import numpy as np
import sys

from .ctext import ctext


def world_to_pixel(inputfile, sky, wave, extnum, verbose=False):
    """Compute world_to_pixel.

    Parameters
    ----------
    inputfile : str, file-like or `pathlib.Path`
        Input FITS filename.
    sky : str or None
        String defining the celestial coordinate with units.
    wave : str or None
        String defining the wavelength with units.
    extnum : int
        Extension number to read the WCS from.
    verbose : bool
        It True, display intermediate information
    """

    with fits.open(inputfile) as hdul:
        if verbose:
            print(hdul.info())
        if extnum > len(hdul):
            raise ValueError(f"Extension number {extnum} exceeds {len(hdul)}")
        header = hdul[extnum].header

    wcs = WCS(header)
    if verbose:
        print(wcs)

    if sky is None or sky == "":
        if wcs.has_celestial:
            sky = wcs.celestial.wcs.crval
            cunit = wcs.celestial.wcs.cunit
            sky = SkyCoord(ra=sky[0]*Unit(cunit[0]), dec=sky[1]*Unit(cunit[1]))
        else:
            sky = None
    else:
        sky = eval(sky)

    if wave is None or wave == "":
        if wcs.has_spectral:
            wave = wcs.spectral.wcs.crval * Unit(wcs.spectral.wcs.cunit[0])
        else:
            wave = None
    else:
        wave = eval(wave)

    print(f'\nComputing pixel for world coordinate {sky}, {wave}:')

    if wcs.has_celestial and wcs.has_spectral:
        result = wcs.world_to_pixel(sky, wave)
    elif wcs.has_celestial:
        result = wcs.world_to_pixel(sky)
    elif wcs.has_spectral:
        result = wcs.world_to_pixel(wave)
    else:
        result = None
    if result is not None:
        result = [item + 1 for item in result]

    output = ''
    for item in result:
        if isinstance(item, np.ndarray):
            output += f"{str(item[0])} "
        else:
            output += f"{str(item)} "
    print(output)


def main(args=None):
    """
    Usage example:
    $ numina-world_to_pixel file.fits \
      --sky 'SkyCoord(0*u.arcsec, 0*u.arcsec)'
      --wave '1.9344e-06*u.m'

    The sky and wavelength coordinates must be provided with units.
    """
    parser = argparse.ArgumentParser(
        description="Convert world to pixel coordinates.",
        formatter_class=RawTextHelpFormatter
    )
    parser.add_argument("inputfile", help="Input FITS file", type=str)
    parser.add_argument("--sky", help="Celestial coordinate (string)\n"
                        "e.g. 'SkyCoord(0*u.arcsec, 0*u.arcsec)'", type=str, default=None)
    parser.add_argument("--wave", help="Spectral wavelength (string)\n"
                        "e.g. '1.9344e-06*u.m'", type=str, default=None)
    parser.add_argument("--extnum", help="Extension number (default 0=PRIMARY)", type=int, default=0)
    parser.add_argument("--verbose", help="Display intermediate information", action="store_true")
    parser.add_argument("--echo", help="Display full command line", action="store_true")
    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    extnum = args.extnum
    if extnum < 0:
        raise ValueError('extnum must be >= 0')

    world_to_pixel(
        inputfile=args.inputfile,
        sky=args.sky,
        wave=args.wave,
        extnum=extnum,
        verbose=args.verbose
    )


if __name__ == '__main__':
    main()



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np

from .raise_valueerror import raise_ValueError


def simulate_delta_lines(line_wave, line_flux, nphotons, wavelength_sampling,
                         rng, wmin=None, wmax=None, plots=False, plot_title=None):
    """Simulate spectrum defined from isolated wavelengths.

    Parameters
    ----------
    line_wave : `~astropy.units.Quantity`
        Numpy array (with astropy units) containing the individual
        wavelength of each line.
    line_flux : array_like
        Array-like object containing the individual flux of each line.
    nphotons : int
        Number of photons to be simulated
    wavelength_sampling : str
        Method to sample the wavelength values. Two options are valid:
        - 'random': the wavelengt of each photon is randomly determined
          using the spectrum shape as the density probability function.
        - 'fixed': the wavelength of each photon is exactly determined
          using the spectrum shape as the density probability function.
    rng : `~numpy.random._generator.Generator`
        Random number generator.
    wmin : `~astropy.units.Quantity`
        Minimum wavelength to be considered.
    wmax : `~astropy.units.Quantity`
        Maximum wavelength to be considered.
    plots : bool
        If True, plot input and output results.
    plot_title : str or None
        Plot title. Used only when 'plots' is True.

    Returns
    -------
    simulated_wave : `~astropy.units.Quantity`
        Wavelength of simulated photons.

    """

    line_flux = np.asarray(line_flux)
    if len(line_wave) != len(line_flux):
        raise_ValueError(f"Incompatible array length: 'line_wave' ({len(line_wave)}), 'line_flux' ({len(line_flux)})")

    if np.any(line_flux < 0):
        raise_ValueError(f'Negative line fluxes cannot be handled')

    if not isinstance(line_wave, u.Quantity):
        raise_ValueError(f"Object 'line_wave': {line_wave} is not a Quantity instance")
    wave_unit = line_wave.unit
    if not wave_unit.is_equivalent(u.m):
        raise_ValueError(f"Unexpected unit for 'line_wave': {wave_unit}")

    # lower wavelength limit
    if wmin is not None:
        if not isinstance(wmin, u.Quantity):
            raise_ValueError(f"Object 'wmin': {wmin} is not a Quantity instance")
        if not wmin.unit.is_equivalent(u.m):
            raise_ValueError(f"Unexpected unit for 'wmin': {wmin}")
        wmin = wmin.to(wave_unit)
        lower_index = np.searchsorted(line_wave.value, wmin.value, side='left')
    else:
        lower_index = 0

    # upper wavelength limit
    if wmax is not None:
        if not isinstance(wmax, u.Quantity):
            raise_ValueError(f"Object 'wmax': {wmax} is not a Quantity instance")
        if not wmax.unit.is_equivalent(u.m):
            raise_ValueError(f"Unexpected unit for 'wmax': {wmin}")
        wmax = wmax.to(wave_unit)
        upper_index = np.searchsorted(line_wave.value, wmax.value, side='right')
    else:
        upper_index = len(line_wave)

    if plots:
        fig, ax = plt.subplots()
        ax.stem(line_wave.value, line_flux, markerfmt=' ', basefmt=' ')
        if wmin is not None:
            ax.axvline(wmin.value, linestyle='--', color='gray')
        if wmax is not None:
            ax.axvline(wmax.value, linestyle='--', color='gray')
        ax.set_xlabel(f'Wavelength ({wave_unit})')
        ax.set_ylabel('Intensity (arbitrary units)')
        if plot_title is not None:
            ax.set_title(plot_title)
        plt.tight_layout()
        plt.show()

    line_wave = line_wave[lower_index:upper_index]
    line_flux = line_flux[lower_index:upper_index]

    # normalized cumulative sum
    cumsum = np.cumsum(line_flux)
    cumsum /= cumsum[-1]

    if plots:
        fig, ax = plt.subplots()
        ax.plot(line_wave.value, cumsum, '-')
        if wmin is not None:
            ax.axvline(wmin.value, linestyle='--', color='gray')
        if wmax is not None:
            ax.axvline(wmax.value, linestyle='--', color='gray')
        ax.set_xlabel(f'Wavelength ({wave_unit})')
        ax.set_ylabel('Cumulative sum')
        if plot_title is not None:
            ax.set_title(plot_title)
        plt.tight_layout()
        plt.show()

    if wavelength_sampling == 'random':
        # samples following a uniform distribution
        unisamples = rng.uniform(low=0, high=1, size=nphotons)
    elif wavelength_sampling == 'fixed':
        # constant sampling of distribution function
        unisamples = np.linspace(0, 1, num=nphotons+1)  # generate extra photon
        unisamples = unisamples[:-1]  # remove last photon
    else:
        unisamples = None  # avoid PyCharm warning
        raise_ValueError(f'Unexpected {wavelength_sampling=}')

    # closest array indices in sorted array
    closest_indices = np.searchsorted(cumsum, unisamples, side='right')

    # simulated wavelengths
    simulated_wave = line_wave.value[closest_indices]
    simulated_wave *= wave_unit

    if plots:
        # count number of photons at each tabulated wavelength value
        x_spectrum, y_spectrum = np.unique(simulated_wave, return_counts=True)

        # scale factor to overplot expected spectrum with same total number
        # of photons as the simulated dataset
        factor = np.sum(line_flux) / nphotons

        # overplot expected and simulated spectrum
        fig, ax = plt.subplots()
        ax.stem(line_wave.value, line_flux / factor, markerfmt=' ', basefmt=' ')
        ax.plot(x_spectrum, y_spectrum, '.')
        if wmin is not None:
            ax.axvline(wmin.value, linestyle='--', color='gray')
        if wmax is not None:
            ax.axvline(wmax.value, linestyle='--', color='gray')
        ax.set_xlabel(f'Wavelength ({wave_unit})')
        ax.set_ylabel('Intensity (number of photons)')
        if plot_title is not None:
            ax.set_title(plot_title)
        plt.tight_layout()
        plt.show()

    return simulated_wave



#
# Copyright 2016-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import typing

from typing_extensions import Self

if typing.TYPE_CHECKING:
    from numina.instrument.configorigin import ElementOrigin
    from numina.instrument.hwdevice import DeviceBase


from numina.instrument.hwdevice import HWDevice
from numina.instrument.signal import Signal


class Carrousel(HWDevice):
    def __init__(
        self,
        cid,
        capacity: int,
        origin: "ElementOrigin | None" = None,
        parent: "DeviceBase | None" = None,
    ):
        super().__init__(name=cid, origin=origin, parent=parent)
        # Container is empty
        self._container = [None] * capacity
        self._capacity = capacity
        self._pos = 0
        # object in the current position
        self._current = self._container[self._pos]

        # signals
        self.changed = Signal()
        self.moved = Signal()

    def current(self):
        return self._current

    def pos(self):
        return self._pos

    def put_in_pos(self, obj, pos: int):
        if pos >= self._capacity or pos < 0:
            raise ValueError("position greater than capacity or negative")

        self._container[pos] = obj
        self._current = self._container[self._pos]

    def move_to(self, pos: int):
        if pos >= self._capacity or pos < 0:
            raise ValueError(f"Position {pos:d} out of bounds")

        if pos != self._pos:
            self._pos = pos
            self._current = self._container[self._pos]
            self.changed.emit(self._pos)
        self.moved.emit(self._pos)

    def select(self, name):
        # find pos of object with name
        for idx, item in enumerate(self._container):
            if item:
                if isinstance(item, str):
                    if item == name:
                        return self.move_to(idx)
                elif item.name == name:
                    return self.move_to(idx)
                else:
                    pass
        else:
            raise ValueError(f"No object named {name}")

    @property
    def position(self):
        return self._pos

    @position.setter
    def position(self, pos: int):
        self.move_to(pos)

    def init_config_info(self):
        info = super().init_config_info()
        if self._current:
            if isinstance(self._current, str):
                selected = self._current
            else:
                try:
                    selected = self._current.config_info()
                except AttributeError:
                    selected = self.label
        else:
            selected = self.label
        info["selected"] = selected
        return info

    @property
    def label(self):
        if self._current:
            if isinstance(self._current, str):
                lab = self._current
            else:
                lab = self._current.name
        else:
            lab = "Unknown"

        return lab

    @label.setter
    def label(self, name):
        self.select(name)

    @classmethod
    def from_component(
        cls,
        name: str,
        comp_id: str,
        origin: "ElementOrigin | None" = None,
        parent: "DeviceBase | None" = None,
        properties=None,
        setup=None,
    ) -> Self:
        capacity = 1
        if setup is not None:
            capacity = setup.values["capacity"]

        obj = cls.__new__(cls)
        obj.__init__(comp_id, capacity, origin=origin, parent=parent)
        return obj


class Wheel(Carrousel):
    def __init__(
        self,
        cid,
        capacity,
        origin: "ElementOrigin | None" = None,
        parent: "DeviceBase | None" = None,
    ):
        super().__init__(cid, capacity, origin=origin, parent=parent)

    def turn(self):
        self._pos = (self._pos + 1) % self._capacity
        self._current = self._container[self._pos]
        self.changed.emit(self._pos)
        self.moved.emit(self._pos)



from datetime import datetime

import pytest

from numina.instrument.assembly import (
    find_element,
    ElementEnum,
    find_instrument,
    process_setup,
    process_properties,
    process_components,
)
from numina.instrument.collection import load_paths_store
from numina.instrument.generic import ComponentGeneric


def test_ins_prefix(pattri_ins):

    devpath1 = pattri_ins.get_device("PATTRI.PSU")
    devpath2 = pattri_ins.get_device("PSU")

    assert isinstance(devpath1, ComponentGeneric)

    assert devpath1 is devpath2


def test_all_comps(pattri_ins):

    id_names = ["PSU", "detector", "wheel"]
    comp_names = ["PSU", "Detector_Model1", "Wheel_Model1"]
    comp_uuids = [
        "41fa8884-2ea3-42c3-98a6-a17a29c990cf",
        "510efa03-c266-4202-8a4d-d5799af49d9d",
        "27614307-783f-4c1c-bd9c-165ff9cec1c5",
    ]
    for id_n, c_name, c_uuid in zip(id_names, comp_names, comp_uuids):
        dev = pattri_ins.get_device(id_n)
        assert dev.origin.name == c_name
        assert str(dev.origin.uuid) == c_uuid


@pytest.fixture
def comp_store():
    pkg_paths = ["numina.drps.tests.configs"]
    # pkg_paths = ["pattridrp.instrument.configs"]
    comp_store = load_paths_store(pkg_paths)
    return comp_store


@pytest.mark.parametrize(
    "date_val", ["2015-05-12T03:04:21", datetime(2015, 5, 12, 3, 4, 21)]
)
def test_find_element1(comp_store, date_val):
    etype = ElementEnum.ELEM_COMPONENT
    el = find_element(
        comp_store,
        etype,
        "27614307-783f-4c1c-bd9c-165ff9cec1c5",
        date_val,
        by_key="uuid",
    )
    assert isinstance(el, dict)


@pytest.mark.parametrize(
    "date_val", ["2015-05-12T03:04:21", datetime(2015, 5, 12, 3, 4, 21)]
)
def test_find_instrument(comp_store, date_val):
    uuid_str = "43273e8c-4071-4a73-a6b4-40c2f07cf054"
    el = find_instrument(comp_store, uuid_str, date_val, by_key="uuid")
    assert isinstance(el, dict)


def test_process_setup(comp_store):
    date_str = "2018-12-12T03:04:21"
    uuid_str = "a7358f24-6ce7-4851-a197-d6515e0592f5"
    block = [{"uuid": uuid_str, "id": "test_block"}]
    aa = process_setup(comp_store, setup_block=block, setup_id="test", date=date_str)
    assert aa.name == "test"
    assert aa.origin is None
    assert "a" in aa.values
    assert "b" in aa.values
    assert isinstance(aa.values["a"], dict)
    assert isinstance(aa.values["b"], dict)


def test_process_setup_error1(comp_store):
    date_str = "2018-12-12T03:04:21"
    block = [{"xxx": "invalid"}]
    with pytest.raises(ValueError):
        process_setup(comp_store, setup_block=block, setup_id="test", date=date_str)


def test_process_properties_error1(comp_store):
    date_str = "2018-12-12T03:04:21"
    block = [{"xxx": "invalid", "id": "test_block"}]
    with pytest.raises(ValueError):
        process_properties(comp_store, prop_block=block, date=date_str)


def test_process_components_error1(comp_store):
    date_str = "2018-12-12T03:04:21"
    block = [{"xxx": "invalid"}]
    with pytest.raises(ValueError):
        process_components(comp_store, comp_block=block, date=date_str)


def test_process_properties_error2(comp_store):
    date_str = "2018-12-12T03:04:21"
    block = [{"one_of": [], "id": "test_block"}]
    with pytest.raises(ValueError):
        process_properties(comp_store, prop_block=block, date=date_str)



import pytest

from ..configorigin import ElementOrigin
from ..elements import ElementEnum, ElementBase


@pytest.mark.parametrize(
    "name, value",
    [
        ("instrument", ElementEnum.ELEM_INSTRUMENT),
        ("component", ElementEnum.ELEM_COMPONENT),
        ("properties", ElementEnum.ELEM_PROPERTIES),
        ("setup", ElementEnum.ELEM_SETUP),
    ],
)
def test_from_str(name, value):
    assert ElementEnum.from_str(name) == value


def test_from_str_err():
    with pytest.raises(ValueError):
        assert ElementEnum.from_str("xxxx")


def test_origin():
    origin = ElementOrigin("Org", "87158f24-6cea-4851-a197-d6515e0592f5")
    a = ElementBase("elem")
    a.set_origin(origin)
    assert a.origin == origin



#
# Copyright 2014 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from .enclosed import EnclosedGaussian  # noqa: F401
from .gaussbox import GaussBox  # noqa: F401
from .sumofgauss import sum_of_gaussian_factory  # noqa: F401



#
# Copyright 2015-2019 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Global cache for testing."""

import os
import sys
import tempfile

import numina.util.context as cntx

from astropy.utils import data


def user_cache_dir(appname=None):
    if sys.platform.startswith('java'):
        import platform
        os_name = platform.java_ver()[3][0]
        if os_name.startswith('Windows'):  # "Windows XP", "Windows 7", etc.
            system = 'win32'
        elif os_name.startswith('Mac'):
            system = 'darwin'
        else:
            system = 'linux2'
    else:
        system = sys.platform

    if system == 'darwin':
        path = os.path.expanduser('~/Library/Caches')
        if appname:
            path = os.path.join(path, appname)
    else:
        if 'XDG_CACHE_HOME' not in os.environ.keys():
            path = os.getenv('XDG_CACHE_HOME', os.path.expanduser('~/.cache'))
        else:
            if appname in os.environ['XDG_CACHE_HOME']:
                path = os.environ['XDG_CACHE_HOME'].split(appname)[0]
            else:
                path = os.getenv('XDG_CACHE_HOME',
                                 os.path.expanduser('~/.cache'))
        if appname:
            path = os.path.join(path, appname)
    if not os.path.exists(os.path.join(path, 'astropy')):
        os.makedirs(os.path.join(path, 'astropy'))
    return path


def download_cache(url, cache=True):
    """Get a tempfile from an URL"""
    cache_dir = user_cache_dir('numina')

    with cntx.environ(XDG_CACHE_HOME=cache_dir):
        fs = open(data.download_file(url, cache=cache), 'rb')
        with tempfile.NamedTemporaryFile(delete=False) as fd:
            block = fs.read()
            while block:
                fd.write(block)
                block = fs.read()

        return fd




import argparse

import astropy.io.fits as fits


def add_key_val(keyname, keyval, keytype, filename, extnum):
    """Add/replace FITS key

    Add/replace the key keyname with value keyval of type keytype in filename.

    Parameters:
    ----------
    keyname : str
        FITS Keyword name.
    keyval : str
        FITS keyword value.
    keytype: str
        FITS keyword type: int, float, str or bool.
    filaname : str
        FITS filename.
    extnum : int
        Extension number where the keyword will be inserted. Note that
        the first extension is number 1 (and not zero).

    """

    funtype = {'int': int, 'float': float, 'str': str, 'bool': bool}
    if keytype not in funtype:
        raise ValueError('Undefined keyword type: ', keytype)
    with fits.open(filename, "update") as hdulist:
        hdulist[extnum].header[keyname] = funtype[keytype](keyval)
        print('>>> Inserting ' + keyname + '=' + keyval + ' in ' + filename)


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('filename',
                        help='FITS file name (wildcards accepted)',
                        nargs='+')
    parser.add_argument('keyname',
                        help='Keyword name')
    parser.add_argument('keyval',
                        help='keyword value')
    parser.add_argument('--keytype',
                        help='Keyword type (int, float, str, bool)',
                        default='str',
                        choices=['int', 'float', 'str', 'bool'])
    parser.add_argument('--extnum',
                        help='Extension number '
                        '(first extension is 1 and not 0)',
                        default=1, type=int)
    args = parser.parse_args(args=args)

    extnum = args.extnum - 1
    for f in args.filename:
        add_key_val(args.keyname, args.keyval, args.keytype, f, extnum)


if __name__ == '__main__':

    main()



from ..signal import Signal


def cb1(arg1):
    return arg1


def cb2(arg2):
    return 0


def cb3(arg1, arg2):
    return 0


def test_signal_cb():
    sig = Signal()
    cid1 = sig.connect(cb1)
    cid2 = sig.connect(cb2)

    res = sig.emit(12)
    assert res == [(cid1, 12), (cid2, 0)]


def test_signal_delete():
    sig = Signal()

    cid1 = sig.connect(cb1)
    cid2 = sig.connect(cb2)

    sig.delete(cid1)
    assert len(sig.callbacks) == 1

    res = sig.emit(12)
    assert res == [(cid2, 0)]


def test_signal_raise(capsys):
    sig = Signal()
    sig.connect(cb1)
    # cb3 produces an error
    sig.connect(cb3)
    res = sig.emit(12)
    captured = capsys.readouterr()
    assert res == [(1, 12)]
    assert "TypeError" in captured.err



#
# Copyright 2016-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

#cython: language_level=3

cimport cython
import numpy as np

# maximum number of intersections of two generic pixels
DEF NMAX_CORNERS = 8


cdef struct Polygon:
    int ncorners
    double x[NMAX_CORNERS]
    double y[NMAX_CORNERS]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef Polygon polygon4vert(double x1, double y1,
                          double x2, double y2,
                          double x3, double y3,
                          double x4, double y4):
    cdef Polygon p
    p.ncorners = 4
    p.x[0] = x1
    p.y[0] = y1
    p.x[1] = x2
    p.y[1] = y2
    p.x[2] = x3
    p.y[2] = y3
    p.x[3] = x4
    p.y[3] = y4
    return p


cdef Polygon empty_polygon():
    cdef Polygon p
    p.ncorners = 0
    return p


@cython.nonecheck(False)
cdef inline check_ncorners(Polygon p):
    if p.ncorners == NMAX_CORNERS:
        raise ValueError('Number of corners exceeds maximum value' +
                         str(NMAX_CORNERS))


# Sutherland–Hodgman algorithm implementation in Cython
# https://stackoverflow.com/questions/44765229/
# The original code has been modified to make use of the Polygon structure
# and the auxiliary functions (note that although the original lists are not
# employed, the variable names have been preserved)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef cclippix_area(Polygon subjectPolygon, Polygon clipPolygon):
    cdef:
        int i, ii
        double cp10, cp11, cp20, cp21
        double s0, s1, e0, e1
        double s_in, e_in
        double dc0, dc1
        double dp0, dp1
        double n1, n2, n3
        double xdum, ydum
        double x1, y1, x2, y2
        double area = 0.0

    outputList = subjectPolygon
    cp10 = clipPolygon.x[clipPolygon.ncorners-1]
    cp11 = clipPolygon.y[clipPolygon.ncorners-1]

    for i in range(clipPolygon.ncorners):
        cp20 = clipPolygon.x[i]
        cp21 = clipPolygon.y[i]

        inputList = outputList
        outputList = empty_polygon()
        s0 = inputList.x[inputList.ncorners-1]
        s1 = inputList.y[inputList.ncorners-1]
        s_in = (cp20 - cp10) * (s1 - cp11) - (cp21 - cp11) * (s0 - cp10)
        for ii in range(inputList.ncorners):
            e0 = inputList.x[ii]
            e1 = inputList.y[ii]
            e_in = (cp20 - cp10) * (e1 - cp11) - (cp21 - cp11) * (e0 - cp10)
            if e_in > 0:
                if s_in <= 0:
                    # compute intersection
                    dc0, dc1 = cp10 - cp20, cp11 - cp21
                    dp0, dp1 =  s0 - e0, s1 - e1
                    n1 = cp10 * cp21 - cp11 * cp20
                    n2 = s0 * e1 - s1 * e0
                    n3 = 1.0 / (dc0 * dp1 - dc1 * dp0)
                    xdum = (n1 * dp0 - n2 * dc0) * n3
                    ydum = (n1 * dp1 - n2 * dc1) * n3
                    check_ncorners(outputList)
                    outputList.x[outputList.ncorners] = xdum
                    outputList.y[outputList.ncorners] = ydum
                    outputList.ncorners += 1
                check_ncorners(outputList)
                outputList.x[outputList.ncorners] = e0
                outputList.y[outputList.ncorners] = e1
                outputList.ncorners += 1
            elif s_in > 0:
                # compute intersection
                dc0, dc1 = cp10 - cp20, cp11 - cp21
                dp0, dp1 =  s0 - e0, s1 - e1
                n1 = cp10 * cp21 - cp11 * cp20
                n2 = s0 * e1 - s1 * e0
                n3 = 1.0 / (dc0 * dp1 - dc1 * dp0)
                xdum = (n1 * dp0 - n2 * dc0) * n3
                ydum = (n1 * dp1 - n2 * dc1) * n3
                check_ncorners(outputList)
                outputList.x[outputList.ncorners] = xdum
                outputList.y[outputList.ncorners] = ydum
                outputList.ncorners += 1
            s0, s1, s_in = e0, e1, e_in
        if outputList.ncorners < 1:
            # no intersection: null area
            return 0.0
        cp10, cp11 = cp20, cp21

    # compute area
    x1 = outputList.x[outputList.ncorners-1]
    y1 = outputList.y[outputList.ncorners-1]
    for i in range(outputList.ncorners):
        x2 = outputList.x[i]
        y2 = outputList.y[i]
        area += x1 * y2 - x2 * y1
        x1, y1 = x2, y2
    area /= 2.0

    return area


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def _resample(double [:,:] image2d,
              double [:] xxx, double [:] yyy,
              long [:] ixx, long [:] iyy,
              int naxis1out=0, int naxis2out=0,
              int ioffx=0, int ioffy=0):
    cdef int naxis1, naxis2
    naxis1 = image2d.shape[1]
    naxis2 = image2d.shape[0]
    if naxis1out == 0:
        naxis1out = naxis1
    if naxis2out == 0:
        naxis2out = naxis2

    image2d_rect_np = np.zeros((naxis2out, naxis1out), dtype=np.double)
    cdef double [:,:] image2d_rect = image2d_rect_np  # memoryview

    cdef:
        long k, k4
        int kk
        int jmin, jmax, imin, imax
        int jdum, idum
        double [:] xvertices
        double [:] yvertices
        double xmin, xmax, ymin, ymax
        double xdum, ydum
        double pixel_area

    for k in range(naxis1out * naxis2out):
        k4 = k*4
        xvertices = xxx[(k4):(k4+4)]
        yvertices = yyy[(k4):(k4+4)]
        polygon1 = polygon4vert(xvertices[0], yvertices[0],
                                xvertices[1], yvertices[1],
                                xvertices[2], yvertices[2],
                                xvertices[3], yvertices[3])
        # limits to determine the pixels of the original (distorted) image
        # that can have intersection with polygon1
        xmin = xmax = xvertices[0]
        for kk in range(1,4):
            xdum = xvertices[kk]
            if xmax < xdum:
                xmax = xdum
            elif xmin > xdum:
                xmin = xdum
        ymin = ymax = yvertices[0]
        for kk in range(1,4):
            ydum = yvertices[kk]
            if ymax < ydum:
                ymax = ydum
            elif ymin > ydum:
                ymin = ydum
        jmin = int(xmin + 0.5)
        jmax = int(xmax + 0.5)
        imin = int(ymin + 0.5)
        imax = int(ymax + 0.5)
        # determine intersection of polygon1 with pixels in the original
        # (distorted) image
        pixel_area = 0
        for jdum in range(jmin, jmax + 1):
            if 0 <= jdum < naxis1:
                for idum in range(imin, imax + 1):
                    if 0 <= idum < naxis2:
                        # polygon corresponding to a particular pixel in
                        # the original (distorted) image
                        polygon2 = polygon4vert(jdum - 0.5, idum - 0.5,
                                                jdum + 0.5, idum - 0.5,
                                                jdum + 0.5, idum + 0.5,
                                                jdum - 0.5, idum + 0.5)
                        # add signal according to the area of the intersection
                        pixel_area += cclippix_area(polygon1, polygon2) * \
                            image2d[idum, jdum]
        image2d_rect[iyy[k], ixx[k]] = pixel_area

    return image2d_rect_np



#
# Copyright 2016-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import pytest

from numina.exceptions import NoResultFound
from numina.tests.drptest import create_drp_test
import numina.core

from ..dictdal import BaseDictDAL
from ..stored import ObservingBlock, StoredProduct


@pytest.fixture
def basedictdal():

    test_drps = create_drp_test(['drptest1.yaml'])
    # com_store = asbl.load_panoply_store(test_drps)
    ob_table = {
        2: dict(id=2, instrument='TEST1', mode="mode1", images=[], children=[], parent=None, facts=None),
        3: dict(id=2, instrument='TEST1', mode="mode2", images=[], children=[], parent=None, facts=None),
    }

    prod_table = {
        'TEST1': {
            '225fcaf2-7f6f-49cc-972a-70fd0aee8e96': [
                {'id': 1, 'type': 'DemoType1', 'tags': {}, 'content': {'demo1': 1}, 'ob': 2},
                {'id': 2, 'type': 'DemoType2', 'tags': {'field2': 'A'}, 'content': {'demo2': 2}, 'ob': 14},
                {'id': 3, 'type': 'DemoType2', 'tags': {'field2': 'B'}, 'content': {'demo2': 3}, 'ob': 15}
            ]
        }
    }

    base = BaseDictDAL(test_drps, ob_table, prod_table, req_table={})
    return base


def test_search_instrument_configuration(basedictdal):
    import numina.instrument.generic

    res = basedictdal.search_instrument_configuration(
        'TEST1', '2017-10-01 12:00:00')

    assert isinstance(res, numina.instrument.generic.InstrumentGeneric)

    res = basedictdal.search_instrument_configuration(
        'TEST1', '2017-10-01 12:00:00')

    assert isinstance(res, numina.instrument.generic.InstrumentGeneric)

    with pytest.raises(ValueError):
        basedictdal.search_instrument_configuration(
            'TEST1', '2011-10-01 12:00:00')

    with pytest.raises(KeyError):
        basedictdal.search_instrument_configuration(
            'TEST2', '2011-10-01 12:00:00')


def test_search_instrument_configuration_from_ob(basedictdal):
    import numina.instrument.generic

    ob = numina.core.ObservationResult(mode=None)

    with pytest.raises(KeyError):
        basedictdal.search_instrument_configuration_from_ob(ob)

    ob = numina.core.ObservationResult(mode='TEST1')
    ob.instrument = 'TEST1'
    res = basedictdal.search_instrument_configuration_from_ob(ob)

    assert isinstance(res, numina.instrument.generic.InstrumentGeneric)

    ob = numina.core.ObservationResult(mode='TEST1')
    ob.instrument = 'TEST1'
    ob.profile = 'missing'

    with pytest.raises(ValueError):
        basedictdal.search_instrument_configuration_from_ob(ob)


def test_search_instrument_configuration_from_ob2(basedictdal):
    import numina.instrument.generic

    ob = numina.core.ObservationResult(instrument='TEST1', mode='TEST1')
    ob.profile = '225fcaf2-7f6f-49cc-972a-70fd0aee8e96'

    insconf = basedictdal.search_instrument_configuration_from_ob(ob)
    assert str(insconf.origin.uuid) == '225fcaf2-7f6f-49cc-972a-70fd0aee8e96'


def test_search_instrument_configuration_from_ob3(basedictdal):
    import numina.instrument.generic

    ob = numina.core.ObservationResult(instrument='TEST1', mode='TEST1')
    ob.profile = '225fcaf2-7f6f-49cc-972a-70fd0aee8e96'

    insconf = basedictdal.search_instrument_configuration_from_ob(ob)
    assert str(insconf.origin.uuid) == '225fcaf2-7f6f-49cc-972a-70fd0aee8e96'


def test_search_oblock(basedictdal):

    with pytest.raises(NoResultFound):
        basedictdal.search_oblock_from_id(obsid=1)

    res = basedictdal.search_oblock_from_id(obsid=2)

    assert isinstance(res, ObservingBlock)

    assert res.id == 2
    assert res.instrument == 'TEST1'


def test_search_recipe(basedictdal):
    from numina.core.utils import AlwaysFailRecipe

    with pytest.raises(KeyError):
        basedictdal.search_recipe('FAIL', 'mode1', 'default')

    with pytest.raises(KeyError):
        basedictdal.search_recipe('TEST1', 'mode1', 'default')

    with pytest.raises(KeyError):
        basedictdal.search_recipe('TEST1', 'fail', 'invalid')

    res = basedictdal.search_recipe('TEST1', 'fail', 'default')
    assert isinstance(res, AlwaysFailRecipe)


def test_search_prod_obsid(basedictdal):

    with pytest.raises(NoResultFound):
        basedictdal.search_prod_obsid('FAIL', 1, 'default')

    with pytest.raises(NoResultFound):
        basedictdal.search_prod_obsid('TEST1', 1, 'default')

    res = basedictdal.search_prod_obsid('TEST1', 2, 'default')
    assert isinstance(res, StoredProduct)


def test_search_prod_req_tags1(basedictdal):

    class DemoType1:
        def name(self):
            return "DemoType1"

    req = numina.core.Requirement(DemoType1, description='Demo1 Requirement')
    ins = 'TEST1'
    version = '225fcaf2-7f6f-49cc-972a-70fd0aee8e96'
    tags = {}
    pipeline = 'default'
    res = basedictdal.search_prod_req_tags(req, ins, version, tags, pipeline)
    assert isinstance(res, StoredProduct)
    assert res.id == 1
    assert res.content == {'demo1': 1}

    assert res.tags == {}


def test_search_prod_req_tags2(basedictdal):

    class DemoType2:
        def name(self):
            return "DemoType2"

    req = numina.core.Requirement(DemoType2, description='Demo2 Requirement')
    ins = 'TEST1'
    version = '225fcaf2-7f6f-49cc-972a-70fd0aee8e96'
    tags = {'field2': 'A'}
    pipeline = 'default'
    res = basedictdal.search_prod_req_tags(req, ins, version, tags, pipeline)
    assert isinstance(res, StoredProduct)
    assert res.id == 2
    assert res.content == {'demo2': 2}

    assert res.tags == {'field2': 'A'}


def test_search_prod_req_tags3(basedictdal):

    class DemoType2:
        def name(self):
            return "DemoType2"

    req = numina.core.Requirement(DemoType2, description='Demo2 Requirement')
    ins = 'TEST1'
    version = '225fcaf2-7f6f-49cc-972a-70fd0aee8e96'
    tags = {'field2': 'C'}
    pipeline = 'default'
    with pytest.raises(NoResultFound):
        basedictdal.search_prod_req_tags(req, ins, version, tags, pipeline)


def test_search_prod_req_tags4(basedictdal):
    class DemoType2:
        def name(self):
            return "DemoType2"

    req = numina.core.Requirement(DemoType2, description='Demo2 Requirement')
    ins = 'TEST1'
    version = '225fcaf2-7f6f-49cc-972a-70fd0aee8e96'
    tags = {}
    pipeline = 'default'
    res = basedictdal.search_prod_req_tags(req, ins, version, tags, pipeline)
    assert isinstance(res, StoredProduct)
    assert res.id == 2
    assert res.content == {'demo2': 2}

    assert res.tags == {'field2': 'A'}



#
# Copyright 2008-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy
import pytest

import numina.array._combine as _c


def test_error1():
    """combine method is not valid"""
    data = numpy.array([[1, 2], [1, 2]])
    with pytest.raises(TypeError):
        _c.generic_combine("dum", [data])


def test_error2():
    """input list is empty"""
    with pytest.raises(ValueError):
        _c.generic_combine(_c.mean_method(), [])


def test_error3():
    """images don't have the same shape"""
    data = numpy.array([[1, 2], [1, 2]])
    databig = numpy.zeros((256, 256))
    with pytest.raises(ValueError):
        _c.generic_combine(_c.mean_method(), [data, databig])


def test_error4():
    """incorrect number of masks"""
    data = numpy.array([[1, 2], [1, 2]])
    mask = numpy.array([[True, False], [True, False]])
    with pytest.raises(ValueError):
        _c.generic_combine(_c.mean_method(), [data], masks=[mask, mask])


def test_error5():
    """mask and image have different shape"""
    data = numpy.array([[1, 2], [1, 2]])
    masks = [numpy.array([True, False, True])]
    with pytest.raises(ValueError):
        _c.generic_combine(_c.mean_method(), [data], masks=masks)


def test_error6():
    """output has wrong shape"""
    data = numpy.array([[1, 2], [1, 2]])
    out = numpy.empty((3, 2, 239))
    with pytest.raises(ValueError):
        _c.generic_combine(_c.mean_method(), [data], out_res=out[0], out_var=out[1], out_pix=out[2])


@pytest.mark.parametrize("bad_par", [
    [[1, 2]],
    [1, 2]
])
def test_scales_error(bad_par):
    data = numpy.array([[1, 2], [1, 2]])

    with pytest.raises(ValueError):
        _c.generic_combine(_c.mean_method(), [data], scales=bad_par)


@pytest.mark.parametrize("bad_par", [
    [[1, 2]],
    [1, 2]
])
def test_zeros_error(bad_par):
    data = numpy.array([[1, 2], [1, 2]])
    with pytest.raises(ValueError):
        _c.generic_combine(_c.mean_method(), [data], zeros=bad_par)


@pytest.mark.parametrize("bad_par", [
    [[1, 2]],
    [1, 2]
])
def test_weights_error(bad_par):
    data = numpy.array([[1, 2], [1, 2]])
    with pytest.raises(ValueError):
        _c.generic_combine(_c.mean_method(), [data], weights=bad_par)


def test_combine_mask_average():
    """Average combine: combination of integer arrays with masks."""
    input1 = numpy.array([[1, 2, 3, 4], [1, 2, 3, 4], [9, 2, 0, 4]])
    input2 = numpy.array([[3, 2, 8, 4], [6, 2, 0, 4], [1, 3, 3, 4]])
    input3 = numpy.array([[7, 2, 1, 4], [1, 2, 0, 4], [44, 2, 2, 0]])
    inputs = [input1, input2, input3]

    mask1 = numpy.array([[False, False, False, True],
                         [False, True, False, False],
                         [False, True, False, False]])
    mask2 = numpy.array([[False, False, False, False],
                         [False, True, False, False],
                         [False, True, False, False]])
    mask3 = numpy.array([[False, False, False, False],
                         [False, True, False, False],
                         [False, False, True, False]])
    masks = [mask1, mask2, mask3]
    rres = numpy.array([[3.66666667, 2., 4., 4.],
                        [2.6666666666666665, 0., 1., 4.],
                        [18., 2., 1.5, 2.66666667]])
    rvar = numpy.array([[3 * 3.11111111, 0., 3 * 4.33333333, 0.],
                        [3 * 2.77777778, 0., 3 * 1.,  0.],
                        [3 * 174.33333333, 0., 2 * 2.25, 3 * 1.77777778]
                        ])

    rnum = numpy.array([[3, 3, 3, 2],
                        [3, 0, 3, 3],
                        [3, 1, 2, 3]])

    out0, out1, out2 = _c.generic_combine(_c.mean_method(), inputs, masks=masks)
    assert numpy.allclose(out0, rres)
    assert numpy.allclose(out1, rvar)
    assert numpy.allclose(out2, rnum)


def test_combine_average():
    """Average combine: combination of integer arrays."""
    input1 = numpy.array([[1, 2, 3, 4], [1, 2, 3, 4], [9, 2, 0, 4]])
    input2 = numpy.array([[3, 2, 8, 4], [6, 2, 0, 4], [1, 3, 3, 4]])
    input3 = numpy.array([[7, 2, 1, 4], [1, 2, 0, 4], [44, 2, 2, 0]])
    inputs = [input1, input2, input3]

    # Results
    rres = numpy.array([[3.66666667, 2., 4., 4.0],
                        [2.6666666666666665, 2., 1., 4.],
                        [18., 2.33333333, 1.666666667, 2.66666667]])
    rvar = 3 * numpy.array([[9.3333333333333339, 0., 13.0, 0.],
                            [8.3333333333333339, 0., 3.00000000, 0.],
                            [523.0, 0.33333333333333337, 2.333333333333333,
                             5.3333333333333339]]) / len(inputs)
    rnum = numpy.array([[3, 3, 3, 3], [3, 3, 3, 3], [3, 3, 3, 3]])

    out0, out1, out2 = _c.generic_combine(_c.mean_method(), inputs)
    assert numpy.allclose(out0, rres)
    assert numpy.allclose(out1, rvar)
    assert numpy.allclose(out2, rnum)


def test_combine_sum():
    """Average combine: combination of integer arrays."""
    # Inputs
    input1 = numpy.array([[1, 2, 3, 4], [1, 2, 3, 4], [9, 2, 0, 4]])
    input2 = numpy.array([[2, 2, 8, 4], [6, 2, 0, 4], [1, 3, 3, 4]])
    input3 = numpy.array([[1, 2, 1, 4], [1, 2, 0, 4], [44, 2, 2, 0]])
    inputs = [input1, input2, input3]
    # Results
    rres = numpy.array([[4, 6, 12, 12],
                        [8, 6, 3, 12],
                        [54, 7, 5, 8]])
    # The variance result is not useful

    rvar = numpy.array([[1, 0, 39, 0],
                        [25, 0, 9, 0],
                        [1569, 1, 7, 16]]) * 3
    rnum = numpy.array([[3, 3, 3, 3], [3, 3, 3, 3], [3, 3, 3, 3]])

    out0, out1, out2 = _c.generic_combine(_c.sum_method(), inputs)
    assert numpy.allclose(out0, rres)
    assert numpy.allclose(out1, rvar)
    assert numpy.allclose(out2, rnum)


def test_combine_median1():
    """Median combine: combination of integer arrays."""
    # Inputs
    input1 = numpy.array([[1, 2, 3, 4], [1, 2, 3, 4], [9, 2, 0, 4]])
    input2 = numpy.array([[1, 2, 3, 4], [1, 2, 3, 4], [9, 2, 0, 4]])
    input3 = numpy.array([[7, 2, 3, 4], [1, 2, 3, 4], [9, 2, 0, 4]])
    input4 = numpy.array([[7, 2, 3, 4], [1, 2, 3, 4], [9, 2, 0, 4]])
    input5 = numpy.array([[7, 2, 1, 4], [1, 2, 0, 4], [44, 2, 2, 0]])
    inputs = [input1, input2, input3, input4, input5]

    rres = input3
    rvar = [[16.954474097331239, 0.0, 1.2558869701726849, 0.0],
            [0.0, 0.0, 2.8257456828885403, 0.0],
            [384.61538461538458, 0.0, 1.2558869701726847,
            5.0235478806907397]]

    out0, out1, out2 = _c.generic_combine(_c.median_method(), inputs)

    assert numpy.allclose(out0, rres)
    assert numpy.allclose(out1, rvar)
    assert numpy.allclose(out2, len(inputs))


def test_combine_median2():
    """Median combine: combination an even number of integer arrays."""
    # Inputs
    input1 = numpy.array([[1, 2, 3, -4]])
    input2 = numpy.array([[1, 2, 6, 4]])
    input3 = numpy.array([[7, 3, 8, -4]])
    input4 = numpy.array([[7, 2, 3, 4]])
    inputs = [input1, input2, input3, input4]

    rres = numpy.array([[4, 2, 4.5, 0.0]], dtype='float')
    rvar = [18.838304552590266, 0.39246467817896391, 9.419152276295133,
            33.490319204604916]

    out0, out1, out2 = _c.generic_combine(_c.median_method(), inputs)

    assert numpy.allclose(out0, rres)
    assert numpy.allclose(out1, rvar)
    assert numpy.allclose(out2, len(inputs))


@pytest.mark.xfail
def test_combine_maxargs():
    """Testing numpy max args limit (with 1 iter impl)"""
    # Inputs
    input1 = numpy.array([[1, 2, 3, -4]])
    inputs = [input1] * 100

    out0, out1, out2 = _c.generic_combine_1iter(_c.median_method(), inputs)
    assert numpy.allclose(out0, input1)


def test_combine_miter_maxargs():
    """Testing numpy max args limit"""
    # Inputs
    input1 = numpy.array([[1, 2, 3, -4]])
    inputs = [input1] * 100

    out0, out1, out2 = _c.generic_combine(_c.median_method(), inputs)
    assert numpy.allclose(out0, input1)


def test_combine_miter_maxargs2():
    """Testing numpy max args limit with larger images"""
    # Inputs
    input1 = numpy.zeros((512, 512))
    inputs = [input1] * 100

    out0, out1, out2 = _c.generic_combine(_c.median_method(), inputs)
    assert numpy.allclose(out0, input1)




import pytest
import astropy.io.fits as fits
import numpy

from ..utils import Combine
from numina.core.oresult import ObservationResult
from numina.tests.simpleobj import create_simple_frame


@pytest.mark.parametrize(
    "method,result", [
        ('mean', 5200.0), ('median', 2800.0), ('minmax', 2666.6667),
        ('sigmaclip', 5200.0)
    ])
def test_combine1(method, result):

    obresult = ObservationResult()
    values = [2000.0, 3000.0, 2200.0, 2800.0, 16000.0]
    nimages = len(values)
    obresult.frames = [
        create_simple_frame(value=value) for value in values
    ]

    recipe = Combine()

    r_input = recipe.create_input(
        obresult=obresult,
        method=method
    )

    r_result = recipe(r_input)
    assert isinstance(r_result, Combine.RecipeResult)

    hdul = r_result.result.open()

    assert isinstance(hdul, fits.HDUList)
    assert numpy.allclose(hdul[0].data, [[result]])
    assert hdul[0].header['NUM-NCOM'] == nimages


@pytest.mark.parametrize(
    "method,kwargs,result", [
        ('minmax', {'nmin': 1, 'nmax': 1}, 2624.8),
        ('minmax', {'nmin': 2, 'nmax': 2}, 2667.0),
        ('minmax', {'nmin': 3, 'nmax': 3}, 2801),
        ('sigmaclip', {'high': 1.0, 'low': 2.0}, 2170.6667)
    ])
def test_combine2(method, kwargs, result):

    obresult = ObservationResult()
    values = [-100, 3123, 2000.0, 3000.0, 2200.0, 2801.0, 16000.0]
    nimages = len(values)
    obresult.frames = [
        create_simple_frame(value=value) for value in values
    ]

    recipe = Combine()

    r_input = recipe.create_input(
        obresult=obresult,
        method=method,
        method_kwargs=kwargs
    )

    r_result = recipe(r_input)
    assert isinstance(r_result, Combine.RecipeResult)

    hdul = r_result.result.open()

    assert isinstance(hdul, fits.HDUList)

    assert numpy.allclose(hdul[0].data, [[result]])

    assert hdul[0].header['NUM-NCOM'] == nimages




import datetime

import astropy.io.fits as fits
import numpy


import numina.types.qc as qc
from ..datamodel import DataModel


def create_test_data():
    return numpy.ones((10, 10), dtype='int32')


def create_test_image(hdr=None):
    data = create_test_data()
    hdu = fits.PrimaryHDU(data)

    hdr = {} if hdr is None else hdr
    for key, val in hdr.items():
        hdu.header[key] = val

    img = fits.HDUList([hdu])
    return img


def test_datamodel1():
    datamodel = DataModel()
    assert datamodel.name == 'UNKNOWN'
    datamodel = DataModel('CLODIA')
    assert datamodel.name == 'CLODIA'


def test_datamodel2():
    img = create_test_image()
    testdata = create_test_data()
    datamodel = DataModel('CLODIA')
    data = datamodel.get_data(img)

    assert numpy.allclose(data, testdata)


def test_qc():
    img = create_test_image()

    datamodel = DataModel('CLODIA')

    qcontrol = datamodel.get_quality_control(img)

    assert qcontrol == qc.QC.UNKNOWN


def test_imgid():

    CHECKSUM = 'RfAdUd2cRd9cRd9c'

    hdr = {'CHECKSUM': CHECKSUM}
    img = create_test_image(hdr)

    datamodel = DataModel('CLODIA')

    imgid_chsum = datamodel.get_imgid(img)

    assert imgid_chsum == CHECKSUM


def test_ginfo():

    CHECKSUM = 'RfAdUd2cRd9cRd9c'

    uuid_str = 'b2f3d815-6f59-48e3-bea1-4d1ea1a3abc1'

    hdr = {
        'CHECKSUM': CHECKSUM,
        'instrume': 'CLODIA',
        'object': '',
        'obsmode': 'TEST',
        'numtype': 'test_img',
        'exptime': 560,
        'darktime': 573,
        'uuid': uuid_str,
        'DATE-OBS': '1975-03-31T12:23:45.00',
        'blckuuid': 1,
        'insconf': 'v1'
    }

    date_obs = datetime.datetime(1975, 3, 31, 12, 23, 45)

    ref = {
        'instrument': 'CLODIA',
        'object': '',
        'n_ext': 1,
        'name_ext': ['PRIMARY'],
        'quality_control': qc.QC.UNKNOWN,
        'mode': 'TEST',
        'type': 'test_img',
        'exptime': 560,
        'darktime': 573,
        'uuid': uuid_str,
        'observation_date': date_obs,
        'blckuuid': '1',
        'block_uuid': 1,
        'imgid': uuid_str,
        'insconf': 'v1', 'insconf_uuid': 'v1'
    }

    img = create_test_image(hdr)

    datamodel = DataModel('CLODIA')

    imgid_chsum = datamodel.gather_info_hdu(img)
    assert imgid_chsum == ref



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import pytest
import numpy as np
from numpy.testing import assert_allclose

from ..traces import trace
from ..traces import axis_to_dispaxis
from ..traces import tracing_limits


def test_axis_to_dispaxis():
    assert axis_to_dispaxis(0) == 1
    assert axis_to_dispaxis(1) == 0

    with pytest.raises(ValueError):
        axis_to_dispaxis(2)


@pytest.mark.parametrize("gauss", [1, 0])
def test_trace_simple(benchmark, gauss):
    """Test a simple trace"""
    arr = np.zeros((100, 100))

    arr[47, 45:55] = 10.0
    arr[48, 45:55] = 100.0
    arr[49, 45:55] = 12.0

    if gauss:
        result = np.array(
            [[44., 48.02061133, 33.36466379], [45., 48.02061133, 66.72932758],
             [46., 48.02061133, 100.09399137], [
                 47., 48.02061133, 100.09399137],
             [48., 48., 100.], [49., 48.02061133, 100.09399137],
             [50., 48.02061133, 100.09399137], [
                 51., 48.02061133, 100.09399137],
             [52., 48.02061133, 100.09399137], [
                 53., 48.02061133, 100.09399137],
             [54., 48.02061133, 66.72932758], [55., 48.02061133, 33.36466379]])
    else:
        result = np.empty((12, 3))
        result[:, 0] = np.arange(44, 56)
        result[:, 1] = 48.0056179775
        result[4, 1] = 48.0
        result[:, 2] = 100.00280898876404
        result[[0, 11], 2] = 33.33426966
        result[[1, 10], 2] = 66.66853933
        result[4, 2] = 100.0

    mm = benchmark(trace, arr, 48.0, 48.0, step=1, gauss=gauss)

    assert mm.shape == (12, 3)
    assert_allclose(mm, result)


def test_trace_bug_27():
    """Trace doesn't work with a flat peak"""
    arr = np.zeros((100, 100))
    arr[47:52, 12:90] = 100.0
    mm = trace(arr, 50, 50)
    assert mm.shape[0] >= 1


def helper_lim_min(col, step, hs):
    while col > hs + step:
        col += -1 * step
    return col


def helper_lim_max(col, step, hs, size):
    while col + step + hs < size:
        col += step
    return col


@pytest.mark.parametrize("size", [4096, 4097])
@pytest.mark.parametrize("col", list(range(1998, 2003)))
@pytest.mark.parametrize("step", [1, 2, 3])
@pytest.mark.parametrize("hs", [1, 2, 3])
def test_lower_limit(size, col, step, hs):
    vcalc_min, vcalc_max = tracing_limits(size, col, step, hs)

    vreal_min = helper_lim_min(col, step, hs)
    vreal_max = helper_lim_max(col, step, hs, size)

    assert vcalc_min == vreal_min
    assert vcalc_max == vreal_max



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Generate a 2D or 3D FITS image with a constant value.
"""
import argparse
from astropy.io import fits
import numpy as np

from .add_script_info_to_fits_history import add_script_info_to_fits_history


def genimage(naxis1, naxis2, naxis3, constant, dtype):
    """Generate a 2D or 3D numpy array with a constant value.
    
    Parameters
    ----------
    naxis1 : int
        Number of pixels in the first axis.
    naxis2 : int
        Number of pixels in the second axis.
    naxis3 : int
        Number of pixels in the third axis.
    constant : float
        Constant value for the array.
    dtype : str
        Data type of the output array.
        
    Returns
    -------
    array :  numpy.ndarray
        The generated image array.
    """
    if naxis1 <= 0 or naxis2 <= 0:
        raise ValueError("naxis1 and naxis2 must be greater than 0 for a 2D image.")
    if naxis3 < 0:
        raise ValueError("naxis3 cannot be negative.")

    if naxis3 == 0:
        # Generate a 2D array
        array = np.full((naxis2, naxis1), constant, dtype=dtype)
    else:
        # Generate a 3D array
        array = np.full((naxis3, naxis2, naxis1), constant, dtype=dtype)

    return array


def main(args=None):
    """Main function to parse arguments and call genimage."""
    parser = argparse.ArgumentParser(description="Generate a 2D or 3D FITS image with a constant value.")
    parser.add_argument('output', type=str, help='Output FITS file name.')
    parser.add_argument('--naxis1', help='Number of pixels for NAXIS1', type=int, default=0)
    parser.add_argument('--naxis2', help='Number of pixels for NAXIS2', type=int, default=0)
    parser.add_argument('--naxis3', help='Number of pixels for NAXIS3', type=int, default=0)
    parser.add_argument('--constant', help='Constant value for the array', type=float, default=0.0)
    parser.add_argument("--dtype", 
                        help="Data type of the output image (default: float32)",
                        type=str, 
                        choices=['uint8', 'int8', 'uint16', 'int16', 'uint32', 'int32', 'uint64', 'int64',
                                 'float32', 'float64'],
                        default='float32')
    parser.add_argument('--overwrite', help='Overwrite the output file if it exists', action='store_true')

    args = parser.parse_args()

    array = genimage(
        args.naxis1, 
        args.naxis2, 
        args.naxis3, 
        args.constant,
        args.dtype
    )

    # save output file
    hdu = fits.PrimaryHDU(array) 
    add_script_info_to_fits_history(hdu.header, args)
    hdu.writeto(args.output, overwrite=args.overwrite)


if __name__ == "__main__":
    main()



#
# Copyright 2014-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from .absdal import AbsDAL  # noqa: F401
from .daliface import DALInterface  # noqa: F401
from .stored import StoredParameter  # noqa: F401
from .stored import StoredProduct  # noqa: F401
from .stored import ObservingBlock  # noqa: F401



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Compute RSS image from detector image"""
from joblib import Parallel, delayed
import numpy as np
import time

from numina.instrument.simulation.ifu.define_3d_wcs import get_wvparam_from_wcs3d
from numina.tools.ctext import ctext

from .update_image2d_rss_method1 import update_image2d_rss_method1


def compute_image2d_rss_from_detector_method1(
        image2d_detector_method0,
        naxis1_detector,
        naxis1_ifu,
        nslices,
        dict_ifu2detector,
        wcs3d,
        noparallel_computation=False,
        verbose=False
):
    """
    Compute the RSS image from the detector image using method 1.
    
    Parameters
    ----------
    image2d_detector_method0 : numpy.ndarray
        The input detector image.
    naxis1_detector : `~astropy.units.Quantity`
        Number of pixels in the first axis of the detector image
        (dispersion direction)
    naxis1_ifu : `~astropy.units.Quantity`
        Number of pixels in the first axis of the IFU image.
    nslices : int
        Number of slices in the IFU image.
    dict_ifu2detector : dict
        Dictionary mapping IFU slices to detector pixels.
    wcs3d : `~astropy.wcs.WCS`
        WCS object for the 3D image.
    noparallel_computation : bool, optional
        If True, do not use parallel computation. Default is False.
    verbose : bool, optional
        If True, print verbose output. Default is False.

    Returns
    -------
    image2d_rss_method1 : numpy.ndarray
        The computed RSS image.
    """

    if verbose:
        print(ctext('\n* Computing image2d RSS (method 1)', fg='green'))

    # get WCS parameters
    wv_cunit1, wv_crpix1, wv_crval1, wv_cdelt1 = get_wvparam_from_wcs3d(wcs3d)

    # initialize image
    image2d_rss_method1 = np.zeros((naxis1_ifu.value * nslices, naxis1_detector.value))

    if verbose:
        print('Rectifying...')
    t0 = time.time()
    if noparallel_computation:
        # explicit loop in slices
        for islice in range(nslices):
            if verbose:
                print(f'{islice=}')
            update_image2d_rss_method1(
                islice=islice,
                image2d_detector_method0=image2d_detector_method0,
                dict_ifu2detector=dict_ifu2detector,
                naxis1_detector=naxis1_detector,
                naxis1_ifu=naxis1_ifu,
                wv_crpix1=wv_crpix1,
                wv_crval1=wv_crval1,
                wv_cdelt1=wv_cdelt1,
                image2d_rss_method1=image2d_rss_method1,
                debug=False
            )
    else:
        Parallel(n_jobs=-1, prefer="threads")(
            delayed(update_image2d_rss_method1)(
                islice=islice,
                image2d_detector_method0=image2d_detector_method0,
                dict_ifu2detector=dict_ifu2detector,
                naxis1_detector=naxis1_detector,
                naxis1_ifu=naxis1_ifu,
                wv_crpix1=wv_crpix1,
                wv_crval1=wv_crval1,
                wv_cdelt1=wv_cdelt1,
                image2d_rss_method1=image2d_rss_method1,
                debug=False
            ) for islice in range(nslices))

    t1 = time.time()
    if verbose:
        print(f'Delta time: {t1 - t0}')

    return image2d_rss_method1



#
# Copyright 2018-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import pytest

from numina.tests.drptest import create_drp_test
import numina.instrument.assembly as asb

from ..dictdal import HybridDAL


@pytest.fixture
def hybriddal():

    drps = create_drp_test(['drpclodia.yaml'])
    name = 'CLODIA'
    ob_table = [
        dict(id=1, instrument=name, mode="sky", images=[], children=[],
             enabled=False
             ),
        dict(id=2, instrument=name, mode="sky", images=[], children=[]),
        dict(id=3, instrument=name, mode="image", images=[], children=[]),
        dict(id=4, instrument=name, mode="sky", images=[], children=[]),
        dict(id=5, instrument=name, mode="image", images=[], children=[]),
        dict(id=30, instrument=name, mode="mosaic", images=[], children=[2, 3]),
        dict(id=40, instrument=name, mode="mosaic", images=[], children=[4, 5]),
        dict(id=400, instrument=name, mode="raiz",
             images=[], children=[30, 40]),
    ]

    prod_table = {
        'TEST1': {
            '225fcaf2-7f6f-49cc-972a-70fd0aee8e96': [
                {'id': 1, 'type': 'DemoType1', 'tags': {}, 'content': {'demo1': 1}, 'ob': 2},
                {'id': 2, 'type': 'DemoType2', 'tags': {'field2': 'A'}, 'content': {'demo2': 2}, 'ob': 14},
                {'id': 3, 'type': 'DemoType2', 'tags': {'field2': 'B'}, 'content': {'demo2': 3}, 'ob': 15}
            ]
        }
    }

    gentable = {}
    gentable['products'] = prod_table
    gentable['requirements'] = {}
    # Load instrument profiles
    pkg_paths = ['numina.drps.tests.configs']
    store = asb.load_paths_store(pkg_paths)
    base = HybridDAL(drps, ob_table, gentable, {}, components=store)

    return base


def test_skip_reserved(hybriddal):

    ss_ids = list(hybriddal.search_session_ids())

    assert ss_ids == [2, 3, 4, 5, 30, 40]


def test_parent_inserted(hybriddal):

    obsres = hybriddal.search_oblock_from_id(2)
    print('OBSR', obsres.__dict__)
    assert obsres.parent == 30

    obsres = hybriddal.search_oblock_from_id(4)
    assert obsres.parent == 40

    obsres = hybriddal.search_oblock_from_id(30)
    assert obsres.parent == 400

    obsres = hybriddal.search_oblock_from_id(400)
    assert obsres.parent is None


def test_previous_obsid(hybriddal):

    obsres = hybriddal.search_oblock_from_id(5)
    previd = hybriddal.search_previous_obsres(obsres, node='prev')
    assert list(previd) == [4, 3, 2, 1]

    obsres = hybriddal.search_oblock_from_id(5)
    previd = hybriddal.search_previous_obsres(obsres, node='prev-rel')
    assert list(previd) == [4]

    obsres = hybriddal.search_oblock_from_id(4)
    previd = hybriddal.search_previous_obsres(obsres, node='prev-rel')
    assert list(previd) == []



import pytest
from astropy.io import fits

from ..keydef import KeyDefinition


def test_get_val1(pattri_header):
    key_def = KeyDefinition(key="INSMODE")
    val = key_def.get_value(pattri_header)
    assert val == "Mode_B"


@pytest.mark.parametrize("convert, value", [(str.upper, "MODE_B"), (None, "Mode_B")])
def test_get_val2(pattri_header, convert, value):
    hdu = fits.PrimaryHDU()
    hdu.header = pattri_header
    hdul = fits.HDUList([hdu])
    key_def = KeyDefinition(key="INSMODE", convert=convert)
    val = key_def.get_value(hdul)
    assert val == value


@pytest.mark.parametrize(
    "value, convert, res",
    [
        ("MODE_C", str.lower, "mode_c"),
        (None, str.lower, "mode_0"),
        ("MODE_C", None, "MODE_C"),
        (None, None, "MODE_0"),
    ],
)
def test_set_val_convert(pattri_header, value, convert, res):
    hdu = fits.PrimaryHDU()
    hdu.header = pattri_header
    hdul = fits.HDUList([hdu])
    key_def = KeyDefinition(key="INSMODE", default="MODE_0", convert=convert)
    key_def.set_value(hdul, value)

    val = key_def.get_value(hdul)
    assert val == res


def test_get_header():
    key_def = KeyDefinition(key="INSMODE", default="MODE_0")
    with pytest.raises(ValueError):
        key_def._get_header({})  # noqa



#
# Copyright 2015-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Compute the wavelength calibration of a particular spectrum."""

import argparse
from astropy.io import fits
import numpy as np
import os
from scipy import ndimage
import sys

from .arccalibration import arccalibration
from .arccalibration import fit_list_of_wvfeatures
from .arccalibration import refine_arccalibration
from ..display.pause_debugplot import pause_debugplot
from ..display.ximplotxy import ximplotxy
from ..stats import robust_std
from ...tools.arg_file_is_new import arg_file_is_new
from .peaks_spectrum import find_peaks_spectrum
from .peaks_spectrum import refine_peaks_spectrum

from ..display.pause_debugplot import DEBUGPLOT_CODES


def collapsed_spectrum(fitsfile, ns1, ns2,
                       method='mean', nwin_background=0,
                       reverse=False, out_sp=None, debugplot=0):
    """Compute a collapsed spectrum from a 2D image using scans in [ns1,ns2].

    Parameters
    ----------
    fitsfile : file object
        File name of FITS file containing the spectra to be calibrated.
    ns1 : int
        First scan (from 1 to NAXIS2).
    ns2 : int
        Last scan (from 1 to NAXIS2).
    method : string
        Indicates collapsing method. Possible values are "mean" or
        "median".
    nwin_background : int
        Window size for the computation of background using a median
        filtering with that window width. This background is computed
        and subtracted only if this parameter is > 0.
    reverse : bool
        If True, reserve wavelength direction prior to wavelength
        calibration.
    out_sp : string or None
        File name to save the selected spectrum in FITS format before
        performing the wavelength calibration.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    sp : 1d numpy array
        Collapsed spectrum.

    """

    # read FITS file
    with fits.open(fitsfile) as hdulist:
        image2d = hdulist[0].data
    naxis2, naxis1 = image2d.shape
    if abs(debugplot) >= 10:
        print('>>> Reading file:', fitsfile.name)
        print('>>> NAXIS1:', naxis1)
        print('>>> NAXIS2:', naxis2)

    if 1 <= ns1 <= ns2 <= naxis2:
        # extract spectrum
        if method == "mean":
            sp = np.mean(image2d[(ns1 - 1):ns2], axis=0)
        elif method == "median":
            sp = np.median(image2d[(ns1 - 1):ns2], axis=0)
        else:
            raise ValueError("Invalid method '" + str(method) + "'")

        # reverse spectrum if necessary
        if reverse:
            sp = sp[::-1]

        # fit and subtract background
        if nwin_background > 0:
            background = ndimage.filters.median_filter(
                sp, size=nwin_background
            )
            sp -= background

        # save spectrum before wavelength calibration in external
        # FITS file
        if out_sp is not None:
            hdu = fits.PrimaryHDU(sp)
            hdu.writeto(out_sp, overwrite=True)
    else:
        raise ValueError("Invalid ns1=" + str(ns1) + ", ns2=" + str(ns2) +
                         " values")

    return sp


def read_wv_master_from_array(master_table, lines='brightest', debugplot=0):
    """read arc line wavelengths from numpy array

    Parameters
    ----------
    master_table : Numpy array
        Numpy array containing the wavelength database.
    lines : string
        Indicates which lines to read. For files with a single column
        or two columns this parameter is irrelevant. For files with
        three columns, lines='brightest' indicates that only the
        brightest lines are read, whereas lines='all' means that all
        the lines are considered.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    wv_master : 1d numpy array
        Array with arc line wavelengths.

    """

    # protection
    if lines not in ['brightest', 'all']:
        raise ValueError('Unexpected lines=' + str(lines))

    # determine wavelengths according to the number of columns
    if master_table.ndim == 1:
        wv_master = master_table
    else:
        wv_master_all = master_table[:, 0]
        if master_table.shape[1] == 2:  # assume old format
            wv_master = np.copy(wv_master_all)
        elif master_table.shape[1] == 3:  # assume new format
            if lines == 'brightest':
                wv_flag = master_table[:, 1]
                wv_master = wv_master_all[np.where(wv_flag == 1)]
            else:
                wv_master = np.copy(wv_master_all)
        else:
            raise ValueError('Lines_catalog file does not have the '
                             'expected number of columns')

    if abs(debugplot) >= 10:
        print("Reading master table from numpy array")
        print("wv_master:\n", wv_master)

    return wv_master


def read_wv_master_file(wv_master_file, lines='brightest', debugplot=0):
    """read arc line wavelengths from external file.

    Parameters
    ----------
    wv_master_file : string
        File name of txt file containing the wavelength database.
    lines : string
        Indicates which lines to read. For files with a single column
        or two columns this parameter is irrelevant. For files with
        three columns, lines='brightest' indicates that only the
        brightest lines are read, whereas lines='all' means that all
        the lines are considered.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    wv_master : 1d numpy array
        Array with arc line wavelengths.

    """

    # protection
    if lines not in ['brightest', 'all']:
        raise ValueError('Unexpected lines=' + str(lines))

    # read table from txt file
    master_table = np.genfromtxt(wv_master_file)

    wv_master = read_wv_master_from_array(master_table, lines)

    if abs(debugplot) >= 10:
        print("Reading master table: " + wv_master_file)
        print("wv_master:\n", wv_master)

    return wv_master


def find_fxpeaks(sp,
                 times_sigma_threshold,
                 minimum_threshold,
                 nwinwidth_initial,
                 nwinwidth_refined,
                 npix_avoid_border,
                 nbrightlines,
                 sigma_gaussian_filtering,
                 minimum_gaussian_filtering,
                 plottitle=None,
                 geometry=None,
                 debugplot=0):
    """Locate line peaks in array coordinates (from 0 to naxis1-1).

    Parameters
    ----------
    sp : 1d numpy array
        Spectrum to be wavelength calibrated.
    times_sigma_threshold : float
        Times robust sigma above the median to detect line peaks.
    minimum_threshold : float or None
            Minimum value of the threshold.
    nwinwidth_initial : int
        Width of the window where each peak must be found using
        the initial method (approximate)
    nwinwidth_refined : int
        Width of the window where each peak location will be
        refined.
    npix_avoid_border : int
            Number of pixels at the borders of the spectrum where peaks
            are not considered. If zero, the actual number will be
            given by nwinwidth_initial.
    nbrightlines : int or list of integers
        Maximum number of brightest lines to be employed in the
        wavelength calibration. If this value is 0, all the detected
        lines will be employed.
    sigma_gaussian_filtering : float
        Sigma of the gaussian filter to be applied to the spectrum in
        order to avoid problems with saturated lines. This filtering is
        skipped when this parameter is <= 0.
    minimum_gaussian_filtering : float
        Minimum pixel value to employ gaussian filtering. This value is
        employed only when sigma_gaussian_filtering is > 0.
    plottile : string
        Plot title.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the Qt backend geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    fxpeaks : 1d numpy array
        Refined location of peaks in array index scale, i.e, from 0
        to naxis1 - 1.
    sxpeaks : 1d numpy array
        Line peak widths.

    """

    # spectrum dimension
    naxis1 = sp.shape[0]

    # apply gaussian filtering when requested
    if sigma_gaussian_filtering > 0:
        spf = ndimage.filters.gaussian_filter(
            sp,
            sigma=sigma_gaussian_filtering
        )
        lpreserve = sp < minimum_gaussian_filtering
        spf[lpreserve] = sp[lpreserve]
    else:
        spf = np.copy(sp)

    # threshold to search for peaks
    q50 = np.percentile(spf, q=50)
    sigma_g = robust_std(spf)
    threshold = q50 + times_sigma_threshold * sigma_g
    if abs(debugplot) >= 10:
        print("median....:", q50)
        print("robuts std:", sigma_g)
        print("threshold.:", threshold)
    if minimum_threshold > threshold:
        threshold = minimum_threshold
    if abs(debugplot) >= 10:
        print("minimum threshold:", minimum_threshold)
        print("final threshold..:", threshold)

    # initial location of the peaks (integer values)
    ixpeaks = find_peaks_spectrum(spf,
                                  nwinwidth=nwinwidth_initial,
                                  threshold=threshold)

    # select a maximum number of brightest lines in each region
    if len(nbrightlines) == 1 and nbrightlines[0] == 0:
        pass
    else:
        if abs(debugplot) >= 10:
            print('nbrightlines =', nbrightlines)
            print('ixpeaks in whole spectrum:\n', ixpeaks)
        region_size = (naxis1-1)/len(nbrightlines)
        ixpeaks_filtered = np.array([], dtype=int)
        for iregion, nlines_in_region in enumerate(nbrightlines):
            if nlines_in_region > 0:
                imin = int(iregion * region_size)
                imax = int((iregion + 1) * region_size)
                if iregion > 0:
                    imin += 1
                ixpeaks_region = \
                    ixpeaks[np.logical_and(ixpeaks >= imin, ixpeaks <= imax)]
                if len(ixpeaks_region) > 0:
                    peak_fluxes = spf[ixpeaks_region]
                    spos = peak_fluxes.argsort()
                    ixpeaks_tmp = ixpeaks_region[spos[-nlines_in_region:]]
                    ixpeaks_tmp.sort()  # in-place sort
                    if abs(debugplot) >= 10:
                        print('ixpeaks in region........:\n', ixpeaks_tmp)
                    ixpeaks_filtered = np.concatenate((ixpeaks_filtered,
                                                       ixpeaks_tmp))
        ixpeaks = ixpeaks_filtered
        if abs(debugplot) >= 10:
            print('ixpeaks filtered.........:\n', ixpeaks)

    # remove peaks too close to any of the borders of the spectrum
    if npix_avoid_border > 0:
        lok_ini = ixpeaks >= npix_avoid_border
        lok_end = ixpeaks <= naxis1 - 1 - npix_avoid_border
        ixpeaks = ixpeaks[lok_ini * lok_end]

    # refined location of the peaks (float values)
    fxpeaks, sxpeaks = refine_peaks_spectrum(spf, ixpeaks,
                                             nwinwidth=nwinwidth_refined,
                                             method="gaussian")

    # print peak location and width of fitted lines
    if abs(debugplot) >= 10:
        print(">>> Number of lines found:", len(fxpeaks))
        print("# line_number, channel, width")
        for i, (fx, sx) in enumerate(zip(fxpeaks, sxpeaks)):
            print(i, fx+1, sx)

    # display median spectrum and peaks
    if abs(debugplot) % 10 != 0:
        xplot = np.arange(1, naxis1 + 1, dtype=float)
        ax = ximplotxy(xplot, sp, show=False, geometry=geometry,
                       **{'label': 'original spectrum'})
        ax.set_xlabel('pixel (from 1 to NAXIS1)')
        ax.set_ylabel('counts')
        if plottitle is not None:
            ax.set_title(plottitle)
        if sigma_gaussian_filtering > 0:
            ax.plot(xplot, spf, label="filtered spectrum")
        ymin = sp.min()
        ymax = sp.max()
        dy = ymax - ymin
        ymin -= dy/20.
        ymax += dy/20.
        ax.set_ylim([ymin, ymax])
        # display threshold
        ax.axhline(y=threshold, color="black", linestyle="dotted",
                   label="detection threshold")
        # mark peak location
        ax.plot(ixpeaks + 1, spf[ixpeaks], 'bo', label="initial location")
        ax.plot(fxpeaks + 1, spf[ixpeaks], 'go', label="refined location")
        # legend
        ax.legend(numpoints=1)
        # show plot
        pause_debugplot(debugplot, pltshow=True)

    return fxpeaks, sxpeaks


def wvcal_spectrum(sp, fxpeaks, poly_degree_wfit, wv_master,
                   wv_ini_search=None, wv_end_search=None,
                   wvmin_useful=None, wvmax_useful=None,
                   geometry=None, debugplot=0):
    """Execute wavelength calibration of a spectrum using fixed line peaks.

    Parameters
    ----------
    sp : 1d numpy array
        Spectrum to be wavelength calibrated.
    fxpeaks : 1d numpy array
        Refined location of peaks in array index scale, i.e, from 0
        to naxis1 - 1. The wavelength calibration is performed using
        these line locations.
    poly_degree_wfit : int
        Degree for wavelength calibration polynomial.
    wv_master : 1d numpy array
        Array with arc line wavelengths.
    wv_ini_search : float or None
        Minimum expected wavelength in spectrum.
    wv_end_search : float or None
        Maximum expected wavelength in spectrum.
    wvmin_useful : float or None
        If not None, this value is used to clip detected lines below it.
    wvmax_useful : float or None
        If not None, this value is used to clip detected lines above it.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the Qt backend geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    solution_wv : instance of SolutionArcCalibration
        Wavelength calibration solution.

    """

    # check there are enough lines for fit
    if len(fxpeaks) <= poly_degree_wfit:
        print(">>> Warning: not enough lines to fit spectrum")
        return None

    # spectrum dimension
    naxis1 = sp.shape[0]

    wv_master_range = wv_master[-1] - wv_master[0]
    delta_wv_master_range = 0.20 * wv_master_range
    if wv_ini_search is None:
        wv_ini_search = wv_master[0] - delta_wv_master_range
    if wv_end_search is None:
        wv_end_search = wv_master[-1] + delta_wv_master_range

    # use channels (pixels from 1 to naxis1)
    xchannel = fxpeaks + 1.0

    # wavelength calibration
    list_of_wvfeatures = arccalibration(
        wv_master=wv_master,
        xpos_arc=xchannel,
        naxis1_arc=naxis1,
        crpix1=1.0,
        wv_ini_search=wv_ini_search,
        wv_end_search=wv_end_search,
        wvmin_useful=wvmin_useful,
        wvmax_useful=wvmax_useful,
        error_xpos_arc=3,
        times_sigma_r=3.0,
        frac_triplets_for_sum=0.50,
        times_sigma_theil_sen=10.0,
        poly_degree_wfit=poly_degree_wfit,
        times_sigma_polfilt=10.0,
        times_sigma_cook=10.0,
        times_sigma_inclusion=10.0,
        geometry=geometry,
        debugplot=debugplot
    )

    title = "Wavelength calibration"
    solution_wv = fit_list_of_wvfeatures(
        list_of_wvfeatures=list_of_wvfeatures,
        naxis1_arc=naxis1,
        crpix1=1.0,
        poly_degree_wfit=poly_degree_wfit,
        weighted=False,
        plot_title=title,
        geometry=geometry,
        debugplot=debugplot
    )

    if abs(debugplot) % 10 != 0:
        # final plot with identified lines
        xplot = np.arange(1, naxis1 + 1, dtype=float)
        ax = ximplotxy(xplot, sp, title=title, show=False,
                       xlabel='pixel (from 1 to NAXIS1)',
                       ylabel='number of counts',
                       geometry=geometry)
        ymin = sp.min()
        ymax = sp.max()
        dy = ymax-ymin
        ymin -= dy/20.
        ymax += dy/20.
        ax.set_ylim([ymin, ymax])
        # plot wavelength of each identified line
        for feature in solution_wv.features:
            xpos = feature.xpos
            reference = feature.reference
            ax.text(xpos, sp[int(xpos+0.5)-1],
                    str(reference), fontsize=8,
                    horizontalalignment='center')
        # show plot
        print('Plot with identified lines')
        pause_debugplot(12, pltshow=True)

    # return the wavelength calibration solution
    return solution_wv


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser()
    # required parameters
    parser.add_argument("fitsfile",
                        help="FITS image containing the spectra",
                        type=argparse.FileType('rb'))
    parser.add_argument("--scans", required=True,
                        help="Tuple ns1[,ns2] (from 1 to NAXIS2)")
    parser.add_argument("--wv_master_file", required=True,
                        help="TXT file containing wavelengths")
    parser.add_argument("--degree", required=True,
                        help="Polynomial degree", type=int)
    # optional arguments
    parser.add_argument("--wvmin",
                        help="Minimum expected wavelength",
                        type=float)
    parser.add_argument("--wvmax",
                        help="Maximum expected wavelength",
                        type=float)
    parser.add_argument("--wvmin_useful",
                        help="Minimum useful wavelength",
                        type=float)
    parser.add_argument("--wvmax_useful",
                        help="Maximum useful wavelength",
                        type=float)
    parser.add_argument("--nwin_background",
                        help="window to compute background (0=none)"
                        " (default=0)",
                        default=0, type=int)
    parser.add_argument("--method",
                        help="collapsing method (default='mean', 'median')",
                        default='mean',
                        choices=['mean', 'median'])
    parser.add_argument("--times_sigma_threshold",
                        help="Threshold (times robust sigma to detect lines)"
                             " (default=10)",
                        default=10, type=float)
    parser.add_argument("--minimum_threshold",
                        help="Minimum threshold to detect lines"
                             " (default=0)",
                        default=0, type=float)
    parser.add_argument("--nwinwidth_initial",
                        help="Initial window width to detect lines"
                             " (default=7)",
                        default=7, type=int)
    parser.add_argument("--nwinwidth_refined",
                        help="Refined window width to detect lines"
                             " (default=5)",
                        default=5, type=int)
    parser.add_argument("--npix_avoid_border",
                        help="Number of pixels in the borders to be avoided"
                             " (default=6)",
                        default=6, type=int)
    parser.add_argument("--nbrightlines",
                        help="Tuple n1,[n2,[n3,...]] with maximum number of "
                             "brightest lines to be used [0=all] (default=0)",
                        default=0)
    parser.add_argument("--degree_refined",
                        help="Degree of the refined fit using faint lines "
                             "from wv_master_file (default=None, i.e. no "
                             "refinement)",
                        default=None, type=int)
    parser.add_argument("--sigma_gauss_filt",
                        help="Sigma (pixels) of gaussian filtering to avoid "
                             "saturared lines (default=0)",
                        default=0, type=float)
    parser.add_argument("--minimum_gauss_filt",
                        help="Minimum pixel value to use gaussian filtering "
                             "(default=0)",
                        default=0, type=float)
    parser.add_argument("--reverse",
                        help="Reverse wavelength direction",
                        action="store_true")
    parser.add_argument("--out_sp",
                        help="File name to save the selected spectrum in FITS "
                             "format before performing the wavelength "
                             "calibration (default=None)",
                        default=None,
                        type=argparse.FileType('wb'))
    parser.add_argument("--geometry",
                        help="tuple x,y,dx,dy (default 0,0,640,480)",
                        default="0,0,640,480")
    parser.add_argument("--pdffile",
                        help="Output PDF file name",
                        type=lambda x: arg_file_is_new(parser, x, mode='wb'))
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting/debugging" +
                        " (default=0)",
                        default=0, type=int,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args=args)

    if args.echo:
        print('\033[1m\033[31m% ' + ' '.join(sys.argv) + '\033[0m\n')

    # ---

    # read pdffile
    if args.pdffile is not None:
        from matplotlib.backends.backend_pdf import PdfPages
        pdf = PdfPages(args.pdffile.name)
        interactive_refinement = False
    else:
        pdf = None
        interactive_refinement = True

    # if refinement is going to be used, check that the corresponding
    # polynomial degree is at least as large as the initial degree
    if args.degree_refined is not None:
        if args.degree > args.degree_refined:
            raise ValueError("degree_refined must be >= degree")

    # scan range
    tmp_str = args.scans.split(",")
    if len(tmp_str) == 2:
        ns1 = int(tmp_str[0])
        ns2 = int(tmp_str[1])
    elif len(tmp_str) == 1:
        ns1 = int(tmp_str[0])
        ns2 = ns1
    else:
        raise ValueError("Invalid tuple for scan range")

    # geometry
    if args.geometry is None:
        geometry = None
    else:
        tmp_str = args.geometry.split(",")
        x_geom = int(tmp_str[0])
        y_geom = int(tmp_str[1])
        dx_geom = int(tmp_str[2])
        dy_geom = int(tmp_str[3])
        geometry = x_geom, y_geom, dx_geom, dy_geom

    # convert nbrightlines in a list of integers
    nbrightlines = [int(nlines) for nlines in args.nbrightlines.split(",")]

    # compute collapsed spectrum
    sp = collapsed_spectrum(
        fitsfile=args.fitsfile, ns1=ns1, ns2=ns2,
        method=args.method,
        nwin_background=args.nwin_background,
        reverse=args.reverse,
        debugplot=args.debugplot
    )

    # read arc line wavelengths from external file (all the lines)
    wv_master_all = read_wv_master_file(args.wv_master_file, lines='all')
    # read arc line wavelengths from external file (brightest lines only)
    wv_master = read_wv_master_file(args.wv_master_file, lines='brightest')

    # clip master arc line lists to expected wavelength range
    if args.wvmin is None:
        wvmin = -np.inf
    else:
        wvmin = args.wvmin
    if args.wvmax is None:
        wvmax = np.inf
    else:
        wvmax = args.wvmax

    lok1 = wvmin <= wv_master_all
    lok2 = wv_master_all <= wvmax
    lok = lok1 * lok2
    if abs(args.debugplot) >= 10:
        print("Number of lines in wv_master_all........: ", len(wv_master_all))
    wv_master_all = wv_master_all[lok]
    if abs(args.debugplot) >= 10:
        print("Number of lines in clipped wv_master_all: ", len(wv_master_all))
        print("clipped wv_master_all:\n", wv_master_all)

    lok1 = wvmin <= wv_master
    lok2 = wv_master <= wvmax
    lok = lok1 * lok2
    if abs(args.debugplot) >= 10:
        print("Number of lines in wv_master............: ", len(wv_master))
    wv_master = wv_master[lok]
    if abs(args.debugplot) >= 10:
        print("Number of lines in clipped wv_master....: ", len(wv_master))
        print("clipped wv_master....:\n", wv_master)

    # determine refined peak location in array coordinates, i.e.,
    # from 0 to (naxis - 1)
    plottitle0 = ' [{}, {}:{}],  line list: {}'.format(
        args.method, ns1, ns2, os.path.basename(args.wv_master_file))
    plottitle = os.path.basename(args.fitsfile.name) + plottitle0
    fxpeaks, sxpeaks = find_fxpeaks(
        sp=sp,
        times_sigma_threshold=args.times_sigma_threshold,
        minimum_threshold=args.minimum_threshold,
        nwinwidth_initial=args.nwinwidth_initial,
        nwinwidth_refined=args.nwinwidth_refined,
        npix_avoid_border=args.npix_avoid_border,
        nbrightlines=nbrightlines,
        sigma_gaussian_filtering=args.sigma_gauss_filt,
        minimum_gaussian_filtering=args.minimum_gauss_filt,
        plottitle=plottitle,
        geometry=geometry,
        debugplot=args.debugplot
    )

    # perform wavelength calibration
    solution_wv = wvcal_spectrum(
        sp=sp,
        fxpeaks=fxpeaks,
        poly_degree_wfit=args.degree,
        wv_master=wv_master,
        wv_ini_search=args.wvmin,
        wv_end_search=args.wvmax,
        wvmin_useful=args.wvmin_useful,
        wvmax_useful=args.wvmax_useful,
        geometry=geometry,
        debugplot=args.debugplot
    )

    # apply gaussian filtering
    if args.sigma_gauss_filt > 0:
        spf = ndimage.filters.gaussian_filter(
            sp,
            sigma=args.sigma_gauss_filt
        )
    else:
        spf = np.copy(sp)

    # save fitted spectrum
    if args.out_sp is not None:
        hdu = fits.PrimaryHDU(spf.astype(np.float32))
        hdu.writeto(args.out_sp, overwrite=True)

    # clip master arc line lists to useful wavelength range
    if args.wvmin_useful is None:
        wvmin = -np.inf
    else:
        wvmin = args.wvmin_useful
    if args.wvmax_useful is None:
        wvmax = np.inf
    else:
        wvmax = args.wvmax_useful

    lok1 = wvmin <= wv_master_all
    lok2 = wv_master_all <= wvmax
    lok = lok1 * lok2
    if abs(args.debugplot) >= 10:
        print("Number of lines in wv_master_all........: ", len(wv_master_all))
    wv_master_all = wv_master_all[lok]
    if abs(args.debugplot) >= 10:
        print("Number of lines in clipped wv_master_all: ", len(wv_master_all))
        print("clipped wv_master_all:\n", wv_master_all)

    # refine wavelength calibration when requested
    if args.degree_refined is not None:
        poly_refined, yres_summary = refine_arccalibration(
            sp=spf,
            poly_initial=np.polynomial.Polynomial(solution_wv.coeff),
            wv_master=wv_master_all,
            poldeg=args.degree_refined,
            ntimes_match_wv=1,
            plottitle=plottitle,
            interactive=interactive_refinement,
            geometry=geometry,
            pdf=pdf,
            debugplot=args.debugplot
        )

    if pdf is not None:
        pdf.close()
    else:
        try:
            input("\nPress RETURN to QUIT...")
        except SyntaxError:
            pass


if __name__ == "__main__":

    main()



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Add script information to the FITS history of a file."""

from astropy.io import fits
from datetime import datetime
from io import BufferedReader
from pathlib import Path
import platform
import sys


def add_script_info_to_fits_history(header, args):
    """Add script information to the FITS header history.

    Parameters
    ----------
    header : `astropy.io.fits.Header`
        The FITS header to which the script information will be added.
    args : `argparse.Namespace`
        The arguments parsed from the command line.

    Returns
    -------
    None
    """
    if not isinstance(header, fits.Header):
        raise ValueError("The header must be an instance of astropy.io.fits.Header.")

    header.add_history('-' * 25)
    header.add_history(f"Date: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")
    header.add_history('-' * 25)
    header.add_history(f'Node: {platform.uname().node}')
    header.add_history(f'Python: {sys.executable}')
    header.add_history(f'$ {Path(sys.argv[0]).name}')
    for arg, value in vars(args).items():
        # filename read as argparse.FileType()
        if isinstance(value, BufferedReader):
            value = value.name if hasattr(value, 'name') else str(value)
        header.add_history(f'--{arg} {value}')



#
# Copyright 2010-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import abc
import logging


_logger = logging.getLogger(__name__)


class Node(metaclass=abc.ABCMeta):
    """An elemental operation in a Flow."""

    def __init__(self, ninputs=1, noutputs=1):
        super(Node, self).__init__()
        self._nin = ninputs
        self._nout = noutputs

    @property
    def ninputs(self):
        return self._nin

    @property
    def noutputs(self):
        return self._nout

    @abc.abstractmethod
    def run(self, img):
        raise NotImplementedError

    def __call__(self, img):
        self.obtain_tuple(img)
        return self.run(img)

    def obtain_tuple(self, arg):
        if isinstance(arg, tuple):
            return arg
        return (arg,)

    def execute(self, arg):
        return self.run(arg)


class AdaptorNode(Node):
    """A :class:`Node` that runs a function."""

    def __init__(self, work, ninputs=1, noutputs=1):
        """work is a function object"""
        super(AdaptorNode, self).__init__(ninputs, noutputs)
        self.work = work

    def run(self, img):
        return self.work(img)


class IdNode(Node):
    """A Node that returns its inputs."""

    def __init__(self):
        """Identity"""
        super(IdNode, self).__init__()

    def run(self, img):
        return img


class OutputSelector(Node):
    """A Node that returns part of the results."""

    def __init__(self, ninputs, indexes):
        noutputs = len(indexes)
        super(OutputSelector, self).__init__(ninputs, noutputs)
        self.indexes = indexes

    def run(self, arg):
        res = tuple(ar for idx, ar in enumerate(arg) if idx in self.indexes)
        if len(res) == 1:
            return res[0]
        return res
